@@ -23,6 +23,8 @@
 
  */
 
+#include <config.h>
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -423,6 +425,10 @@ static DFBResult driver_open_device(CoreInputDevice *device,
   }
 
   data = D_CALLOC(1, sizeof(eloData));
+  if (!data) {
+    close(fd);
+    return D_OOM();
+  }
 
   data->fd     = fd;
   data->device = device;
