@@ -1,5 +1,5 @@
 /*
-   (c) Copyright 2001-2007  The DirectFB Organization (directfb.org)
+   (c) Copyright 2001-2008  The world wide DirectFB Open Source Community (directfb.org)
    (c) Copyright 2000-2004  Convergence (integrated media) GmbH
 
    All rights reserved.
@@ -29,6 +29,7 @@
 #include <config.h>
 
 #include <sys/types.h>
+#include <alloca.h>
 #include <dirent.h>
 #include <string.h>
 #include <stdio.h>
@@ -79,9 +80,9 @@ suppress_module (const char *name)
           if (strcmp (direct_config->disable_module[i], name) == 0) {
                D_INFO( "Direct/Modules: suppress module '%s'\n", direct_config->disable_module[i] );
                return 1;
-	  }
+          }
 
-	  i++;
+          i++;
      }
 
      return 0;
@@ -99,10 +100,12 @@ direct_modules_register( DirectModuleDir *directory,
      D_ASSERT( name != NULL );
      D_ASSERT( funcs != NULL );
 
-     D_DEBUG_AT( Direct_Modules, "Registering '%s' ('%s')\n", name, directory->path );
+     D_DEBUG_AT( Direct_Modules, "Registering '%s' ('%s')...\n", name, directory->path );
 
 #ifdef DYNAMIC_LINKING
      if ((entry = lookup_by_name( directory, name )) != NULL) {
+          D_MAGIC_ASSERT( entry, DirectModuleEntry );
+
           entry->loaded = true;
           entry->funcs  = funcs;
 
@@ -110,10 +113,21 @@ direct_modules_register( DirectModuleDir *directory,
      }
 #endif
 
-     if (directory->loading)
+     if (directory->loading) {
           entry = directory->loading;
-     else if (! (entry = D_CALLOC( 1, sizeof(DirectModuleEntry) )))
-          return;
+          D_MAGIC_ASSERT( entry, DirectModuleEntry );
+
+          directory->loading = NULL;
+     }
+     else {
+          entry = D_CALLOC( 1, sizeof(DirectModuleEntry) );
+          if (!entry) {
+               D_OOM();
+               return;
+          }
+
+          D_MAGIC_SET( entry, DirectModuleEntry );
+     }
 
      entry->directory = directory;
      entry->loaded    = true;
@@ -130,13 +144,40 @@ direct_modules_register( DirectModuleDir *directory,
           entry->disabled = true;
      }
 
-     D_MAGIC_SET( entry, DirectModuleEntry );
-
      direct_list_prepend( &directory->entries, &entry->link );
 
      D_DEBUG_AT( Direct_Modules, "...registered.\n" );
 }
 
+void
+direct_modules_unregister( DirectModuleDir *directory,
+                           const char      *name )
+{
+     DirectModuleEntry *entry;
+
+     D_DEBUG_AT( Direct_Modules, "Unregistering '%s' ('%s')...\n", name, directory->path );
+
+#ifdef DYNAMIC_LINKING
+     entry = lookup_by_name( directory, name );
+     if (!entry) {
+          D_ERROR( "Direct/Modules: Unregister failed, could not find '%s' module!\n", name );
+          return;
+     }
+
+     D_MAGIC_ASSERT( entry, DirectModuleEntry );
+
+     D_FREE( entry->name );
+
+     direct_list_remove( &directory->entries, &entry->link );
+
+     D_MAGIC_CLEAR( entry );
+
+     D_FREE( entry );
+#endif
+
+     D_DEBUG_AT( Direct_Modules, "...unregistered.\n" );
+}
+
 int
 direct_modules_explore_directory( DirectModuleDir *directory )
 {
@@ -146,15 +187,31 @@ direct_modules_explore_directory( DirectModuleDir *directory )
      struct dirent *entry = NULL;
      struct dirent  tmp;
      int            count = 0;
+     const char    *pathfront = "";
+     const char    *path;
+     char          *buf;
 
      D_ASSERT( directory != NULL );
      D_ASSERT( directory->path != NULL );
 
-     dir_len = strlen( directory->path );
-     dir     = opendir( directory->path );
+     D_DEBUG_AT( Direct_Modules, "%s( '%s' )\n", __FUNCTION__, directory->path );
+
+     path = directory->path;
+
+     if(path[0]!='/') {
+          pathfront = direct_config->module_dir;
+          if(!pathfront)
+               pathfront = MODULEDIR;
+     }
+
+     buf = alloca( strlen(pathfront) + 1 + strlen(path) + 1 ); /* pre, slash, post, 0 */
+     sprintf( buf, "%s%s%s", pathfront, ( pathfront && path && (path[0] != '/') && (pathfront[strlen(pathfront)-1] != '/') ) ? "/" : "", path );
+
+     dir_len = strlen( buf );
+     dir     = opendir( buf );
 
      if (!dir) {
-          D_PERROR( "Direct/Modules: Could not open module directory `%s'!\n", directory->path );
+          D_DEBUG_AT( Direct_Modules, "  -> ERROR opening directory: %s!\n", strerror(errno) );
           return 0;
      }
 
@@ -176,6 +233,8 @@ direct_modules_explore_directory( DirectModuleDir *directory )
           if (!module)
                continue;
 
+          D_MAGIC_SET( module, DirectModuleEntry );
+
           module->directory = directory;
           module->dynamic   = true;
           module->file      = D_STRDUP( entry->d_name );
@@ -186,7 +245,7 @@ direct_modules_explore_directory( DirectModuleDir *directory )
           if ((handle = open_module( module )) != NULL) {
                if (!module->loaded) {
                     int    len;
-                    void (*func)();
+                    void (*func)( void );
 
                     D_ERROR( "Direct/Modules: Module '%s' did not register itself after loading! "
                              "Trying default module constructor...\n", entry->d_name );
@@ -211,8 +270,6 @@ direct_modules_explore_directory( DirectModuleDir *directory )
                     if (!module->loaded) {
                          module->disabled = true;
 
-                         D_MAGIC_SET( module, DirectModuleEntry );
-
                          direct_list_prepend( &directory->entries,
                                               &module->link );
                     }
@@ -232,8 +289,6 @@ direct_modules_explore_directory( DirectModuleDir *directory )
           else {
                module->disabled = true;
 
-               D_MAGIC_SET( module, DirectModuleEntry );
-
                direct_list_prepend( &directory->entries, &module->link );
           }
 
@@ -364,13 +419,25 @@ unload_module( DirectModuleEntry *module )
 static void *
 open_module( DirectModuleEntry *module )
 {
-     DirectModuleDir *directory = module->directory;
-     int              entry_len = strlen(module->file);
-     int              buf_len   = strlen(directory->path) + entry_len + 2;
-     char             buf[buf_len];
+     DirectModuleDir *directory;
      void            *handle;
+     char            *pathfront;
+     const char      *path;
 
-     snprintf( buf, buf_len, "%s/%s", directory->path, module->file );
+     D_MAGIC_ASSERT( module, DirectModuleEntry );
+
+     directory = module->directory;
+     path      = directory->path;
+     pathfront = "";
+
+     if(path[0]!='/') {
+          pathfront = direct_config->module_dir;
+          if(!pathfront)
+               pathfront = MODULEDIR;
+     }
+
+     char buf[ strlen(pathfront) + 1 + strlen(path) + 1 + strlen(module->file) + 1 ];
+     sprintf( buf, "%s%s%s/%s", pathfront, ( pathfront && path && (path[0] != '/') && (pathfront[strlen(pathfront)-1] != '/') ) ? "/" : "", path, module->file );
 
      D_DEBUG_AT( Direct_Modules, "Loading '%s'...\n", buf );
 
