@@ -1,5 +1,5 @@
 /*
-   (c) Copyright 2001-2007  The DirectFB Organization (directfb.org)
+   (c) Copyright 2001-2008  The world wide DirectFB Open Source Community (directfb.org)
    (c) Copyright 2000-2004  Convergence (integrated media) GmbH
 
    All rights reserved.
@@ -37,7 +37,7 @@
 #include <fusion/reactor.h>
 #include <direct/debug.h>
 
-typedef void (*FusionObjectDestructor)( FusionObject *object, bool zombie );
+typedef void (*FusionObjectDestructor)( FusionObject *object, bool zombie, void *ctx );
 
 typedef bool (*FusionPropIterator)( char *key, void *value, void *ctx);
 
@@ -80,6 +80,7 @@ FusionObjectPool *fusion_object_pool_create ( const char             *name,
                                               int                     object_size,
                                               int                     message_size,
                                               FusionObjectDestructor  destructor,
+                                              void                   *ctx,
                                               const FusionWorld      *world );
 
 DirectResult      fusion_object_pool_destroy( FusionObjectPool       *pool,
@@ -131,6 +132,18 @@ prefix##_attach( type         *object,                                         \
 }                                                                              \
                                                                                \
 static inline DirectResult                                                     \
+prefix##_attach_channel( type         *object,                                 \
+                         int           channel,                                \
+                         ReactionFunc  func,                                   \
+                         void         *ctx,                                    \
+                         Reaction     *ret_reaction )                          \
+{                                                                              \
+     D_MAGIC_ASSERT( (FusionObject*) object, FusionObject );                   \
+     return fusion_reactor_attach_channel( ((FusionObject*)object)->reactor,   \
+                                           channel, func, ctx, ret_reaction ); \
+}                                                                              \
+                                                                               \
+static inline DirectResult                                                     \
 prefix##_detach( type     *object,                                             \
                  Reaction *reaction )                                          \
 {                                                                              \
@@ -170,6 +183,18 @@ prefix##_dispatch( type               *object,                                 \
 }                                                                              \
                                                                                \
 static inline DirectResult                                                     \
+prefix##_dispatch_channel( type               *object,                         \
+                           int                 channel,                        \
+                           void               *message,                        \
+                           int                 size,                           \
+                           const ReactionFunc *globals )                       \
+{                                                                              \
+     D_MAGIC_ASSERT( (FusionObject*) object, FusionObject );                   \
+     return fusion_reactor_dispatch_channel( ((FusionObject*)object)->reactor, \
+                                      channel, message, size, true, globals ); \
+}                                                                              \
+                                                                               \
+static inline DirectResult                                                     \
 prefix##_ref( type *object )                                                   \
 {                                                                              \
      D_MAGIC_ASSERT( (FusionObject*) object, FusionObject );                   \
@@ -204,7 +229,7 @@ prefix##_link( type **link,                                                    \
                                                                                \
      *link = object;                                                           \
                                                                                \
-     return DFB_OK;                                                            \
+     return DR_OK;                                                            \
 }                                                                              \
                                                                                \
 static inline DirectResult                                                     \
