@@ -1,5 +1,5 @@
 /*
-   (c) Copyright 2001-2007  The DirectFB Organization (directfb.org)
+   (c) Copyright 2001-2008  The world wide DirectFB Open Source Community (directfb.org)
    (c) Copyright 2000-2004  Convergence (integrated media) GmbH
 
    All rights reserved.
@@ -60,6 +60,12 @@
 #define IDIRECTFB_METHOD_ID_WaitForSync               27
 #define IDIRECTFB_METHOD_ID_GetInterface              28
 
+typedef struct {
+     int  width;
+     int  height;
+     int  bpp;
+} IDirectFB_Dispatcher_EnumVideoModes_Item;
+
 typedef struct {
      DFBScreenID          screen_id;
      DFBScreenDescription desc;
