@@ -1,5 +1,5 @@
 /*
-   (c) Copyright 2001-2007  The DirectFB Organization (directfb.org)
+   (c) Copyright 2001-2008  The world wide DirectFB Open Source Community (directfb.org)
    (c) Copyright 2000-2004  Convergence (integrated media) GmbH
 
    All rights reserved.
@@ -47,7 +47,7 @@
 #include <core/layer_control.h>
 #include <core/layer_region.h>
 #include <core/screen.h>
-#include <core/surfaces.h>
+#include <core/surface.h>
 #include <core/system.h>
 #include <core/windows.h>
 #include <core/windowstack.h>
@@ -79,8 +79,7 @@ static DFBResult allocate_surface    ( CoreLayer                   *layer,
 
 static DFBResult reallocate_surface  ( CoreLayer                   *layer,
                                        CoreLayerRegion             *region,
-                                       CoreLayerRegionConfig       *config,
-                                       DFBDisplayLayerConfig       *previous );
+                                       CoreLayerRegionConfig       *config );
 
 static DFBResult deallocate_surface  ( CoreLayer                   *layer,
                                        CoreLayerRegion             *region );
@@ -92,7 +91,7 @@ static void      screen_rectangle    ( CoreLayerContext            *context,
 /**********************************************************************************************************************/
 
 static void
-context_destructor( FusionObject *object, bool zombie )
+context_destructor( FusionObject *object, bool zombie, void *ctx )
 {
      CoreLayerContext *context = (CoreLayerContext*) object;
      CoreLayer        *layer   = dfb_layer_at( context->layer_id );
@@ -107,6 +106,8 @@ context_destructor( FusionObject *object, bool zombie )
      /* Remove the context from the layer's context stack. */
      dfb_layer_remove_context( layer, context );
 
+     dfb_layer_context_lock( context );
+
      /* Destroy the window stack. */
      if (context->stack) {
           dfb_windowstack_destroy( context->stack );
@@ -135,7 +136,7 @@ dfb_layer_context_pool_create( const FusionWorld *world )
      return fusion_object_pool_create( "Layer Context Pool",
                                        sizeof(CoreLayerContext),
                                        sizeof(CoreLayerContextNotification),
-                                       context_destructor, world );
+                                       context_destructor, NULL, world );
 }
 
 /**********************************************************************************************************************/
@@ -169,6 +170,7 @@ dfb_layer_context_init( CoreLayerContext *context,
      context->layer_id   = shared->layer_id;
      context->config     = shared->default_config;
      context->adjustment = shared->default_adjustment;
+     context->rotation   = dfb_config->layers[dfb_layer_id_translated(layer)].rotate;
 
      /* Initialize screen location. */
      context->screen.location.x = 0.0f;
@@ -209,6 +211,7 @@ dfb_layer_context_init( CoreLayerContext *context,
 DFBResult
 dfb_layer_context_activate( CoreLayerContext *context )
 {
+     DFBResult        ret;
      int              index;
      CoreLayer       *layer;
      CoreLayerRegion *region;
@@ -238,6 +241,14 @@ dfb_layer_context_activate( CoreLayerContext *context )
           /* Activate each region. */
           if (dfb_layer_region_activate( region ))
                D_WARN( "could not activate region!" );
+
+          if (region->surface) {
+               D_ASSERT( region->surface_lock.buffer == NULL );
+
+               ret = reallocate_surface( layer, region, &region->config );
+               if (ret)
+                    D_DERROR( ret, "Core/Layers: Reallocation of layer surface failed!\n" );
+          }
      }
 
      context->active = true;
@@ -248,8 +259,14 @@ dfb_layer_context_activate( CoreLayerContext *context )
                                             layer->layer_data, &context->adjustment );
 
      /* Resume window stack. */
-     if (context->stack)
-          dfb_wm_set_active( context->stack, true );
+     if (context->stack) {
+          CoreWindowStack *stack = context->stack;
+
+          D_MAGIC_ASSERT( stack, CoreWindowStack );
+
+          if (stack->flags & CWSF_INITIALIZED)
+               dfb_wm_set_active( stack, true );
+     }
 
      /* Unlock the context. */
      dfb_layer_context_unlock( context );
@@ -287,8 +304,14 @@ dfb_layer_context_deactivate( CoreLayerContext *context )
      context->active = false;
 
      /* Suspend window stack. */
-     if (context->stack)
-          dfb_wm_set_active( context->stack, false );
+     if (context->stack) {
+          CoreWindowStack *stack = context->stack;
+
+          D_MAGIC_ASSERT( stack, CoreWindowStack );
+
+          if (stack->flags & CWSF_ACTIVATED)
+               dfb_wm_set_active( stack, false );
+     }
 
      /* Unlock the context. */
      dfb_layer_context_unlock( context );
@@ -321,7 +344,8 @@ dfb_layer_context_add_region( CoreLayerContext *context,
      }
 
      /* Inherit state from context. */
-     region->state = context->active ? CLRSF_ACTIVE : CLRSF_NONE;
+     if (context->active)
+          region->state |= CLRSF_ACTIVE;
 
      /* Unlock the context. */
      dfb_layer_context_unlock( context );
@@ -477,7 +501,7 @@ dfb_layer_context_test_configuration( CoreLayerContext            *context,
      CoreLayer                  *layer;
      CoreLayerRegionConfig       region_config;
      CoreLayerRegionConfigFlags  failed;
-     DisplayLayerFuncs          *funcs;
+     const DisplayLayerFuncs    *funcs;
 
      D_ASSERT( context != NULL );
      D_ASSERT( config != NULL );
@@ -553,12 +577,19 @@ DFBResult
 dfb_layer_context_set_configuration( CoreLayerContext            *context,
                                      const DFBDisplayLayerConfig *config )
 {
+     int                         i;
      DFBResult                   ret;
      CoreLayer                  *layer;
      CoreLayerShared            *shared;
      CoreLayerRegionConfig       region_config;
      CoreLayerRegionConfigFlags  flags;
-     DisplayLayerFuncs          *funcs;
+     const DisplayLayerFuncs    *funcs;
+
+     D_DEBUG_AT( Core_Layers, "%s (%p, %p)\n", __FUNCTION__, context, config );
+
+     D_DEBUG_AT( Core_Layers, "%s (%p, %p)\n", __FUNCTION__, context, config );
+
+     D_DEBUG_AT( Core_Layers, "%s (%p, %p)\n", __FUNCTION__, context, config );
 
      D_ASSERT( context != NULL );
      D_ASSERT( config != NULL );
@@ -618,8 +649,16 @@ dfb_layer_context_set_configuration( CoreLayerContext            *context,
                bool                      surface    = shared->description.caps & DLCAPS_SURFACE;
                CoreLayerRegionStateFlags configured = region->state & CLRSF_CONFIGURED;
 
-               if (shared->description.caps & DLCAPS_SOURCES)
-                    surface = (region_config.source_id == DLSID_SURFACE);
+               if (shared->description.caps & DLCAPS_SOURCES) {
+                    for (i=0; i<shared->description.sources; i++) {
+                         if (shared->sources[i].description.source_id == region_config.source_id)
+                              break;
+                    }
+
+                    D_ASSERT( i < shared->description.sources );
+
+                    surface = shared->sources[i].description.caps & DDLSCAPS_SURFACE;
+               }
 
                D_FLAGS_CLEAR( region->state, CLRSF_CONFIGURED );
 
@@ -627,11 +666,24 @@ dfb_layer_context_set_configuration( CoreLayerContext            *context,
                if (surface) {
                     flags |= CLRCF_SURFACE | CLRCF_PALETTE;
 
-                    if (region->surface)
-                         ret = reallocate_surface( layer, region, &region_config,
-                                                   &context->config );
-                    else
+                    if (region->surface) {
+                         if (D_FLAGS_IS_SET( region->state, CLRSF_REALIZED )) {
+                              if (!D_FLAGS_IS_SET( region->state, CLRSF_FROZEN ))
+                                   D_ASSUME( region->surface_lock.buffer != NULL );
+
+                              if (region->surface_lock.buffer)
+                                   dfb_surface_unlock_buffer( region->surface, &region->surface_lock );
+                         }
+
+                         ret = reallocate_surface( layer, region, &region_config );
+                         if (ret)
+                              D_DERROR( ret, "Core/Layers: Reallocation of layer surface failed!\n" );
+                    }
+                    else {
                          ret = allocate_surface( layer, region, &region_config );
+                         if (ret)
+                              D_DERROR( ret, "Core/Layers: Allocation of layer surface failed!\n" );
+                    }
 
                     if (ret) {
                          dfb_layer_region_unlock( region );
@@ -768,13 +820,15 @@ update_primary_region_config( CoreLayerContext           *context,
 
 DFBResult
 dfb_layer_context_set_src_colorkey( CoreLayerContext *context,
-                                    u8 r, u8 g, u8 b )
+                                    u8 r, u8 g, u8 b, int index )
 {
      DFBResult             ret;
      CoreLayerRegionConfig config;
 
      D_ASSERT( context != NULL );
 
+     D_DEBUG_AT( Core_Layers, "%s (%02x %02x %02x - %d)\n", __FUNCTION__, r, g, b, index );
+
      /* Lock the context. */
      if (dfb_layer_context_lock( context ))
           return DFB_FUSION;
@@ -787,6 +841,9 @@ dfb_layer_context_set_src_colorkey( CoreLayerContext *context,
      config.src_key.g = g;
      config.src_key.b = b;
 
+     if (index >= 0)
+          config.src_key.index = index & 0xff;
+
      /* Try to set the new configuration. */
      ret = update_primary_region_config( context, &config, CLRCF_SRCKEY );
 
@@ -798,7 +855,7 @@ dfb_layer_context_set_src_colorkey( CoreLayerContext *context,
 
 DFBResult
 dfb_layer_context_set_dst_colorkey( CoreLayerContext *context,
-                                    u8 r, u8 g, u8 b )
+                                    u8 r, u8 g, u8 b, int index )
 {
      DFBResult             ret;
      CoreLayerRegionConfig config;
@@ -817,6 +874,9 @@ dfb_layer_context_set_dst_colorkey( CoreLayerContext *context,
      config.dst_key.g = g;
      config.dst_key.b = b;
 
+     if (index >= 0)
+          config.dst_key.index = index & 0xff;
+
      /* Try to set the new configuration. */
      ret = update_primary_region_config( context, &config, CLRCF_DSTKEY );
 
@@ -1029,6 +1089,29 @@ dfb_layer_context_set_opacity( CoreLayerContext *context,
      return ret;
 }
 
+DFBResult
+dfb_layer_context_set_rotation( CoreLayerContext *context,
+                                int               rotation )
+{
+     D_ASSERT( context != NULL );
+
+     /* Lock the context. */
+     if (dfb_layer_context_lock( context ))
+          return DFB_FUSION;
+
+     /* Do nothing if the rotation didn't change. */
+     if (context->rotation != rotation) {
+          context->rotation = rotation;
+
+          dfb_windowstack_repaint_all( context->stack );
+     }
+
+     /* Unlock the context. */
+     dfb_layer_context_unlock( context );
+
+     return DFB_OK;
+}
+
 DFBResult
 dfb_layer_context_set_coloradjustment( CoreLayerContext         *context,
                                        const DFBColorAdjustment *adjustment )
@@ -1169,24 +1252,21 @@ dfb_layer_context_set_clip_regions( CoreLayerContext *context,
 }
 
 DFBResult
-dfb_layer_context_create_window( CoreDFB                 *core,
-                                 CoreLayerContext        *context,
-                                 int                      x,
-                                 int                      y,
-                                 int                      width,
-                                 int                      height,
-                                 DFBWindowCapabilities    caps,
-                                 DFBSurfaceCapabilities   surface_caps,
-                                 DFBSurfacePixelFormat    pixelformat,
-                                 CoreWindow             **ret_window )
+dfb_layer_context_create_window( CoreDFB                     *core,
+                                 CoreLayerContext            *context,
+                                 const DFBWindowDescription  *desc,
+                                 CoreWindow                 **ret_window )
 {
      DFBResult        ret;
      CoreWindow      *window;
      CoreWindowStack *stack;
      CoreLayer       *layer;
 
+     D_DEBUG_AT( Core_Layers, "%s()\n", __FUNCTION__ );
+
      D_ASSERT( context != NULL );
      D_ASSERT( context->stack != NULL );
+     D_ASSERT( desc != NULL );
      D_ASSERT( ret_window != NULL );
 
      layer = dfb_layer_at( context->layer_id );
@@ -1207,8 +1287,7 @@ dfb_layer_context_create_window( CoreDFB                 *core,
           }
      }
 
-     ret = dfb_window_create( stack, x, y, width, height, caps,
-                              surface_caps, pixelformat, &window );
+     ret = dfb_window_create( stack, desc, &window );
      if (ret) {
           dfb_layer_context_unlock( context );
           return ret;
@@ -1349,6 +1428,7 @@ build_updated_config( CoreLayer                   *layer,
      /* Update source and destination rectangle. */
      if (update->flags & (DLCONF_WIDTH | DLCONF_HEIGHT)) {
           int width, height;
+          DFBResult ret;
 
           flags |= CLRCF_SOURCE | CLRCF_DEST;
 
@@ -1359,10 +1439,11 @@ build_updated_config( CoreLayer                   *layer,
 
           switch (context->screen.mode) {
                case CLLM_CENTER:
-                    dfb_screen_get_screen_size( layer->screen, &width, &height );
-
-                    ret_config->dest.x = (width  - ret_config->width)  / 2;
-                    ret_config->dest.y = (height - ret_config->height) / 2;
+                    ret = dfb_screen_get_layer_dimension( layer->screen, layer, &width, &height );
+                    if( ret == DFB_OK ) {
+                         ret_config->dest.x = (width  - ret_config->width)  / 2;
+                         ret_config->dest.y = (height - ret_config->height) / 2;
+                    }
                     /* fall through */
 
                case CLLM_POSITION:
@@ -1422,12 +1503,15 @@ allocate_surface( CoreLayer             *layer,
                   CoreLayerRegion       *region,
                   CoreLayerRegionConfig *config )
 {
-     DFBResult               ret;
-     DisplayLayerFuncs      *funcs;
-     CoreSurface            *surface = NULL;
-     DFBSurfaceCapabilities  caps    = DSCAPS_VIDEOONLY;
+     DFBResult                ret;
+     const DisplayLayerFuncs *funcs;
+     CoreSurface             *surface = NULL;
+     DFBSurfaceCapabilities   caps    = DSCAPS_VIDEOONLY;
+     CoreSurfaceTypeFlags     type    = CSTF_LAYER;
+     CoreSurfaceConfig        scon;
 
      D_ASSERT( layer != NULL );
+     D_ASSERT( layer->shared != NULL );
      D_ASSERT( layer->funcs != NULL );
      D_ASSERT( region != NULL );
      D_ASSERT( region->surface == NULL );
@@ -1451,20 +1535,20 @@ allocate_surface( CoreLayer             *layer,
           }
      }
      else {
+          CoreLayerShared *shared = layer->shared;
+
           /* Choose surface capabilities depending on the buffer mode. */
           switch (config->buffermode) {
                case DLBM_FRONTONLY:
                     break;
 
-               case DLBM_TRIPLE:
-                    caps |= DSCAPS_TRIPLE;
-                    break;
-
                case DLBM_BACKVIDEO:
+               case DLBM_BACKSYSTEM:
                     caps |= DSCAPS_DOUBLE;
                     break;
 
-               case DLBM_BACKSYSTEM:
+               case DLBM_TRIPLE:
+                    caps |= DSCAPS_TRIPLE;
                     break;
 
                default:
@@ -1481,25 +1565,27 @@ allocate_surface( CoreLayer             *layer,
                                           DSCAPS_SEPARATED  |
                                           DSCAPS_PREMULTIPLIED);
 
+          scon.flags  = CSCONF_SIZE | CSCONF_FORMAT | CSCONF_CAPS;
+          scon.size.w = config->width;
+          scon.size.h = config->height;
+          scon.format = config->format;
+          scon.caps   = caps;
+
+          if (shared->contexts.primary == region->context)
+               type |= CSTF_SHARED;
+
           /* Use the default surface creation. */
-          ret = dfb_surface_create( layer->core,
-                                    config->width, config->height,
-                                    config->format, CSP_VIDEOONLY,
-                                    caps, NULL, &surface );
+          ret = dfb_surface_create( layer->core, &scon, type, shared->layer_id, NULL, &surface );
           if (ret) {
                D_DERROR( ret, "Core/layers: Surface creation failed!\n" );
                return ret;
           }
 
-          if (config->buffermode == DLBM_BACKSYSTEM) {
-               surface->caps |= DSCAPS_DOUBLE;
-               ret = dfb_surface_reconfig( surface, CSP_VIDEOONLY, CSP_SYSTEMONLY );
-               if (ret)
-                    D_DERROR( ret, "Core/layers: Surface reconfiguration failed!\n" );
-          }
+          if (config->buffermode == DLBM_BACKSYSTEM)
+               surface->buffers[1]->policy = CSP_SYSTEMONLY;
      }
 
-     /* Tell the region about it's new surface (adds a global reference). */
+     /* Tell the region about its new surface (adds a global reference). */
      ret = dfb_layer_region_set_surface( region, surface );
 
      /* Remove local reference of dfb_surface_create(). */
@@ -1511,12 +1597,12 @@ allocate_surface( CoreLayer             *layer,
 static DFBResult
 reallocate_surface( CoreLayer             *layer,
                     CoreLayerRegion       *region,
-                    CoreLayerRegionConfig *config,
-                    DFBDisplayLayerConfig *previous )
+                    CoreLayerRegionConfig *config )
 {
-     DFBResult          ret;
-     DisplayLayerFuncs *funcs;
-     CoreSurface       *surface;
+     DFBResult                ret;
+     const DisplayLayerFuncs *funcs;
+     CoreSurface             *surface;
+     CoreSurfaceConfig        sconfig;
 
      D_ASSERT( layer != NULL );
      D_ASSERT( layer->funcs != NULL );
@@ -1524,8 +1610,6 @@ reallocate_surface( CoreLayer             *layer,
      D_ASSERT( region->surface != NULL );
      D_ASSERT( config != NULL );
      D_ASSERT( config->buffermode != DLBM_WINDOWS );
-     D_ASSERT( previous != NULL );
-     D_ASSERT( previous->buffermode != DLBM_WINDOWS );
 
      funcs   = layer->funcs;
      surface = region->surface;
@@ -1536,66 +1620,76 @@ reallocate_surface( CoreLayer             *layer,
                                            region->region_data,
                                            config, surface );
 
-     /* FIXME: write surface management functions
-               for easier configuration changes */
+     sconfig.flags = CSCONF_SIZE | CSCONF_FORMAT | CSCONF_CAPS;
 
-     if (previous->buffermode != config->buffermode) {
-          DFBSurfaceCapabilities old_caps = surface->caps;
+     sconfig.caps = surface->config.caps & ~(DSCAPS_FLIPPING  | DSCAPS_INTERLACED |
+                                             DSCAPS_SEPARATED | DSCAPS_PREMULTIPLIED);
 
-          switch (config->buffermode) {
-               case DLBM_TRIPLE:
-                    surface->caps |=  DSCAPS_TRIPLE;
-                    surface->caps &= ~DSCAPS_DOUBLE;
-                    ret = dfb_surface_reconfig( surface,
-                                                CSP_VIDEOONLY, CSP_VIDEOONLY );
-                    break;
-               case DLBM_BACKVIDEO:
-                    surface->caps |=  DSCAPS_DOUBLE;
-                    surface->caps &= ~DSCAPS_TRIPLE;
-                    ret = dfb_surface_reconfig( surface,
-                                                CSP_VIDEOONLY, CSP_VIDEOONLY );
-                    break;
-               case DLBM_BACKSYSTEM:
-                    surface->caps |=  DSCAPS_DOUBLE;
-                    surface->caps &= ~DSCAPS_TRIPLE;
-                    ret = dfb_surface_reconfig( surface,
-                                                CSP_VIDEOONLY, CSP_SYSTEMONLY );
-                    break;
-               case DLBM_FRONTONLY:
-                    surface->caps &= ~DSCAPS_FLIPPING;
-                    ret = dfb_surface_reconfig( surface,
-                                                CSP_VIDEOONLY, CSP_VIDEOONLY );
-                    break;
+     switch (config->buffermode) {
+          case DLBM_TRIPLE:
+               sconfig.caps |= DSCAPS_TRIPLE;
+               break;
 
-               default:
-                    D_BUG("unknown buffermode");
-                    return DFB_BUG;
-          }
+          case DLBM_BACKVIDEO:
+          case DLBM_BACKSYSTEM:
+               sconfig.caps |= DSCAPS_DOUBLE;
+               break;
 
-          if (ret) {
-               surface->caps = old_caps;
-               return ret;
-          }
-     }
+          case DLBM_FRONTONLY:
+               break;
 
-     if (config->width  != previous->width  ||
-         config->height != previous->height ||
-         config->format != previous->pixelformat)
-     {
-          ret = dfb_surface_reformat( layer->core, surface, config->width,
-                                      config->height, config->format );
-          if (ret)
-               return ret;
+          default:
+               D_BUG("unknown buffermode");
+               return DFB_BUG;
      }
 
      /* Add available surface capabilities. */
-     surface->caps &= ~(DSCAPS_INTERLACED | DSCAPS_SEPARATED | DSCAPS_PREMULTIPLIED);
-     surface->caps |= config->surface_caps & (DSCAPS_INTERLACED |
-                                              DSCAPS_SEPARATED  |
-                                              DSCAPS_PREMULTIPLIED); 
-     /* FIXME: remove this? */
+     sconfig.caps |= config->surface_caps & (DSCAPS_INTERLACED |
+                                             DSCAPS_SEPARATED  |
+                                             DSCAPS_PREMULTIPLIED); 
+
      if (config->options & DLOP_DEINTERLACING)
-          surface->caps |= DSCAPS_INTERLACED;
+          sconfig.caps |= DSCAPS_INTERLACED;
+
+     sconfig.size.w = config->width;
+     sconfig.size.h = config->height;
+     sconfig.format = config->format;
+
+     ret = dfb_surface_lock( surface );
+     if (ret)
+          return ret;
+
+     ret = dfb_surface_reconfig( surface, &sconfig );
+     if (ret) {
+          dfb_surface_unlock( surface );
+          return ret;
+     }
+
+     if (DFB_PIXELFORMAT_IS_INDEXED(surface->config.format) && !surface->palette) {
+          ret = dfb_surface_init_palette( layer->core, surface );
+          if (ret)
+               D_DERROR( ret, "Core/Layers: Could not initialize palette while switching to indexed mode!\n" );
+     }
+
+     switch (config->buffermode) {
+          case DLBM_TRIPLE:
+          case DLBM_BACKVIDEO:
+               surface->buffers[1]->policy = CSP_VIDEOONLY;
+               break;
+
+          case DLBM_BACKSYSTEM:
+               surface->buffers[1]->policy = CSP_SYSTEMONLY;
+               break;
+
+          case DLBM_FRONTONLY:
+               break;
+
+          default:
+               D_BUG("unknown buffermode");
+               return DFB_BUG;
+     }
+
+     dfb_surface_unlock( surface );
      
      return DFB_OK;
 }
@@ -1603,9 +1697,9 @@ reallocate_surface( CoreLayer             *layer,
 static DFBResult
 deallocate_surface( CoreLayer *layer, CoreLayerRegion *region )
 {
-     DFBResult          ret;
-     DisplayLayerFuncs *funcs;
-     CoreSurface       *surface;
+     DFBResult                ret;
+     const DisplayLayerFuncs *funcs;
+     CoreSurface             *surface;
 
      D_ASSERT( layer != NULL );
      D_ASSERT( layer->funcs != NULL );
@@ -1652,9 +1746,9 @@ screen_rectangle( CoreLayerContext  *context,
 
      D_ASSERT( layer->screen != NULL );
 
-     ret = dfb_screen_get_screen_size( layer->screen, &width, &height );
+     ret = dfb_screen_get_layer_dimension( layer->screen, layer, &width, &height );
      if (ret) {
-          D_WARN( "could not determine screen size" );
+          D_WARN( "could not determine mixer/screen dimension of layer %d", context->layer_id );
 
           rect->x = location->x * 720;
           rect->y = location->y * 576;
