@@ -1,13 +1,14 @@
 /*
-   (c) Copyright 2000-2002  convergence integrated media GmbH.
-   (c) Copyright 2002-2004  convergence GmbH.
+   (c) Copyright 2001-2008  The world wide DirectFB Open Source Community (directfb.org)
+   (c) Copyright 2000-2004  Convergence (integrated media) GmbH
 
    All rights reserved.
 
    Written by Denis Oliver Kropp <dok@directfb.org>,
               Andreas Hundt <andi@fischlustig.de>,
-              Sven Neumann <neo@directfb.org> and
-              Ville Syrjl <syrjala@sci.fi>.
+              Sven Neumann <neo@directfb.org>,
+              Ville Syrjälä <syrjala@sci.fi> and
+              Claudio Ciccani <klan@users.sf.net>.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -63,6 +64,7 @@
 #include <direct/trace.h>
 #include <direct/util.h>
 
+#include <gfx/convert.h>
 #include <gfx/util.h>
 
 #include <core/layers_internal.h>
@@ -96,30 +98,18 @@ static const ReactionFunc dfb_window_globals[] = {
  * Window destructor.
  */
 static void
-window_destructor( FusionObject *object, bool zombie )
+window_destructor( FusionObject *object, bool zombie, void *ctx )
 {
-     CoreWindow      *window = (CoreWindow*) object;
-     CoreWindowStack *stack  = window->stack;
+     CoreWindow *window = (CoreWindow*) object;
 
      D_DEBUG_AT( Core_Windows, "destroying %p (%d,%d - %dx%d%s)\n", window,
                  DFB_RECTANGLE_VALS( &window->config.bounds ), zombie ? " ZOMBIE" : "");
 
-     D_ASSUME( window->stack != NULL );
+     dfb_window_destroy( window );
 
-     if (stack) {
-          dfb_windowstack_lock( stack );
-
-          if (stack->cursor.window == window)
-               stack->cursor.window = NULL;
-
-          dfb_window_destroy( window );
-
-          /* Unlink the primary region of the context. */
-          if (window->primary_region)
-               dfb_layer_region_unlink( &window->primary_region );
-
-          dfb_windowstack_unlock( stack );
-     }
+     /* Unlink the primary region of the context. */
+     if (window->primary_region)
+          dfb_layer_region_unlink( &window->primary_region );
 
      fusion_object_destroy( object );
 }
@@ -130,7 +120,7 @@ dfb_window_pool_create( const FusionWorld *world )
      return fusion_object_pool_create( "Window Pool",
                                        sizeof(CoreWindow),
                                        sizeof(DFBWindowEvent),
-                                       window_destructor, world );
+                                       window_destructor, NULL, world );
 }
 
 /**************************************************************************************************/
@@ -148,6 +138,7 @@ create_region( CoreDFB                 *core,
      CoreLayerRegionConfig  config;
      CoreLayerRegion       *region;
      CoreSurface           *surface;
+     CoreSurfaceConfig      scon;
 
      D_ASSERT( core != NULL );
      D_ASSERT( context != NULL );
@@ -199,16 +190,20 @@ create_region( CoreDFB                 *core,
                else if (config.options & DLOP_ALPHACHANNEL)
                     config.options = (config.options & ~DLOP_ALPHACHANNEL) | DLOP_OPACITY;
                else {
-                    D_ERROR( "DirectFB/Core/Windows: Unable to set region configuration!\n" );
+                    D_DERROR( ret, "DirectFB/Core/Windows: Unable to set region configuration!\n" );
                     dfb_layer_region_unref( region );
                     return ret;
                }
           }
      } while (ret);
 
+     scon.flags  = CSCONF_SIZE | CSCONF_FORMAT | CSCONF_CAPS;
+     scon.size.w = config.width;
+     scon.size.h = config.height;
+     scon.format = format;
+     scon.caps   = surface_caps | DSCAPS_VIDEOONLY;
 
-     ret = dfb_surface_create( core, config.width, config.height, format,
-                               CSP_VIDEOONLY, surface_caps, NULL, &surface );
+     ret = dfb_surface_create( core, &scon, CSTF_SHARED | CSTF_LAYER, context->layer_id, NULL, &surface );
      if (ret) {
           dfb_layer_region_unref( region );
           return ret;
@@ -235,32 +230,32 @@ create_region( CoreDFB                 *core,
 }
 
 DFBResult
-dfb_window_create( CoreWindowStack        *stack,
-                   int                     x,
-                   int                     y,
-                   int                     width,
-                   int                     height,
-                   DFBWindowCapabilities   caps,
-                   DFBSurfaceCapabilities  surface_caps,
-                   DFBSurfacePixelFormat   pixelformat,
-                   CoreWindow            **ret_window )
+dfb_window_create( CoreWindowStack             *stack,
+                   const DFBWindowDescription  *desc,
+                   CoreWindow                 **ret_window )
 {
-     DFBResult          ret;
-     CoreSurface       *surface;
-     CoreSurfacePolicy  surface_policy = CSP_SYSTEMONLY;
-     CoreLayer         *layer;
-     CoreLayerContext  *context;
-     CoreWindow        *window;
-     CardCapabilities   card_caps;
-     CoreWindowConfig   config;
+     DFBResult               ret;
+     CoreSurface            *surface;
+     CoreSurfacePolicy       surface_policy = CSP_SYSTEMONLY;
+     CoreLayer              *layer;
+     CoreLayerContext       *context;
+     CoreWindow             *window;
+     CardCapabilities        card_caps;
+     CoreWindowConfig        config;
+     DFBWindowCapabilities   caps;
+     DFBSurfaceCapabilities  surface_caps;
+     DFBSurfacePixelFormat   pixelformat;
+
+     D_DEBUG_AT( Core_Windows, "%s( %p )\n", __FUNCTION__, stack );
 
      D_ASSERT( stack != NULL );
      D_ASSERT( stack->context != NULL );
-     D_ASSERT( width > 0 );
-     D_ASSERT( height > 0 );
+     D_ASSERT( desc != NULL );
+     D_ASSERT( desc->width > 0 );
+     D_ASSERT( desc->height > 0 );
      D_ASSERT( ret_window != NULL );
 
-     if (width > 4096 || height > 4096)
+     if (desc->width > 4096 || desc->height > 4096)
           return DFB_LIMITEXCEEDED;
 
      /* Lock the window stack. */
@@ -270,8 +265,13 @@ dfb_window_create( CoreWindowStack        *stack,
      context = stack->context;
      layer   = dfb_layer_at( context->layer_id );
 
-     surface_caps &= DSCAPS_INTERLACED | DSCAPS_SEPARATED | DSCAPS_PREMULTIPLIED |
-                     DSCAPS_DEPTH | DSCAPS_STATIC_ALLOC | DSCAPS_SYSTEMONLY | DSCAPS_VIDEOONLY;
+
+     caps         = desc->caps;
+     pixelformat  = desc->pixelformat;
+     surface_caps = desc->surface_caps & (DSCAPS_INTERLACED    | DSCAPS_SEPARATED  |
+                                          DSCAPS_PREMULTIPLIED | DSCAPS_DEPTH      |
+                                          DSCAPS_STATIC_ALLOC  | DSCAPS_SYSTEMONLY |
+                                          DSCAPS_VIDEOONLY);
 
      if (!dfb_config->translucent_windows) {
           caps &= ~DWCAPS_ALPHACHANNEL;
@@ -329,50 +329,61 @@ dfb_window_create( CoreWindowStack        *stack,
           }
      }
 
-     surface_caps &= ~(DSCAPS_SYSTEMONLY | DSCAPS_VIDEOONLY);
-
-     switch (surface_policy) {
-          case CSP_SYSTEMONLY:
-               surface_caps |= DSCAPS_SYSTEMONLY;
-               break;
-
-          case CSP_VIDEOONLY:
-               surface_caps |= DSCAPS_VIDEOONLY;
-               break;
-
-          default:
-               break;
-     }
+     dfb_surface_caps_apply_policy( surface_policy, &surface_caps );
 
      if (caps & DWCAPS_DOUBLEBUFFER)
           surface_caps |= DSCAPS_DOUBLE;
 
+
      memset( &config, 0, sizeof(CoreWindowConfig) );
 
-     config.bounds.x = x;
-     config.bounds.y = y;
-     config.bounds.w = width;
-     config.bounds.h = height;
+     config.bounds.x = desc->posx;
+     config.bounds.y = desc->posy;
+     config.bounds.w = desc->width;
+     config.bounds.h = desc->height;
+     config.stacking = (desc->flags & DWDESC_STACKING) ? desc->stacking : DWSC_MIDDLE;
 
      config.events   = DWET_ALL;
 
-     /* Auto enable blending for ARGB only, not LUT8. */
+     /* Auto enable blending for ARGB only, not indexed. */
      if ((caps & DWCAPS_ALPHACHANNEL) &&
-         DFB_PIXELFORMAT_HAS_ALPHA(pixelformat) && pixelformat != DSPF_LUT8)
+          DFB_PIXELFORMAT_HAS_ALPHA (pixelformat) &&
+         !DFB_PIXELFORMAT_IS_INDEXED(pixelformat))
           config.options |= DWOP_ALPHACHANNEL;
 
+     /* Override automatic settings. */
+     if (desc->flags & DWDESC_OPTIONS)
+          config.options = desc->options;
+
      /* Create the window object. */
      window = dfb_core_create_window( layer->core );
 
-     window->id     = ++stack->id_pool;
-     window->caps   = caps;
-     window->stack  = stack;
-     window->config = config;
+     window->id        = ++stack->id_pool;
+     window->caps      = caps;
+     window->stack     = stack;
+     window->config    = config;
+     window->parent_id = (desc->flags & DWDESC_PARENT) ? desc->parent_id : 0;
+
+     ret = dfb_wm_preconfigure_window( stack, window );
+     if(ret) {
+          fusion_object_destroy( &window->object );
+          dfb_windowstack_unlock( stack );
+          return ret;
+     }
+
+     /* wm may have changed values */
+     config = window->config;
+     caps   = window->caps;
+
+     if (dfb_config->warn.flags & DCWF_CREATE_WINDOW)
+          D_WARN( "create-window   %4dx%4d %6s, caps 0x%08x, surface-caps 0x%08x, ID %u",
+                  window->config.bounds.w, window->config.bounds.h, dfb_pixelformat_name(pixelformat),
+                  window->caps, surface_caps, window->id );
 
      /* Create the window's surface using the layer's palette if possible. */
      if (! (caps & DWCAPS_INPUTONLY)) {
           if (context->config.buffermode == DLBM_WINDOWS) {
-               CoreLayerRegion *region;
+               CoreLayerRegion *region = NULL;
 
                /* Create a region for the window. */
                ret = create_region( layer->core, context, window,
@@ -390,10 +401,6 @@ dfb_window_create( CoreWindowStack        *stack,
                /* Link the surface into the window structure. */
                dfb_surface_link( &window->surface, surface );
                dfb_surface_unref( surface );
-
-               /* Attach our global listener to the surface. */
-               dfb_surface_attach_global( surface, DFB_WINDOW_SURFACE_LISTENER,
-                                          window, &window->surface_reaction );
           }
           else {
                CoreLayerRegion *region;
@@ -410,36 +417,51 @@ dfb_window_create( CoreWindowStack        *stack,
                dfb_layer_region_link( &window->primary_region, region );
                dfb_layer_region_unref( region );
 
-               /* Create the surface for the window. */
-               ret = dfb_surface_create( layer->core,
-                                         width, height, pixelformat,
-                                         surface_policy, surface_caps,
-                                         region->surface ?
-                                         region->surface->palette : NULL, &surface );
-               if (ret) {
-                    dfb_layer_region_unlink( &window->primary_region );
-                    fusion_object_destroy( &window->object );
-                    dfb_windowstack_unlock( stack );
-                    return ret;
+               D_DEBUG_AT( Core_Windows, "  -> %dx%d %s %s\n",
+                           window->config.bounds.w, window->config.bounds.h,
+                           dfb_pixelformat_name(pixelformat),
+                           (surface_policy == CSP_VIDEOONLY) ?
+                              "VIDEOONLY" :
+                              ((surface_policy == CSP_SYSTEMONLY) ?
+                                   "SYSTEM ONLY" :
+                                   "AUTO VIDEO") );
+
+               /* Give the WM a chance to provide its own surface. */
+               if (!window->surface) {
+                    /* Create the surface for the window. */
+                    ret = dfb_surface_create_simple( layer->core, config.bounds.w, config.bounds.h,
+                                                     pixelformat, surface_caps, CSTF_SHARED | CSTF_WINDOW,
+                                                     (desc->flags & DWDESC_RESOURCE_ID) ?
+                                                     desc->resource_id : window->id,
+                                                     region->surface ?
+                                                     region->surface->palette : NULL, &surface );
+                    if (ret) {
+                         D_DERROR( ret, "Core/Windows: Failed to create window surface!\n" );
+                         dfb_layer_region_unlink( &window->primary_region );
+                         fusion_object_destroy( &window->object );
+                         dfb_windowstack_unlock( stack );
+                         return ret;
+                    }
+
+                    /* Link the surface into the window structure. */
+                    dfb_surface_link( &window->surface, surface );
+                    dfb_surface_unref( surface );
                }
-
-               /* Link the surface into the window structure. */
-               dfb_surface_link( &window->surface, surface );
-               dfb_surface_unref( surface );
-
-               /* Attach our global listener to the surface. */
-               dfb_surface_attach_global( surface, DFB_WINDOW_SURFACE_LISTENER,
-                                          window, &window->surface_reaction );
           }
      }
+     else
+          D_DEBUG_AT( Core_Windows, "  -> %dx%d - INPUT ONLY!\n",
+                      window->config.bounds.w, window->config.bounds.h );
+
+     D_DEBUG_AT( Core_Windows, "  -> %p\n", window );
 
      /* Pass the new window to the window manager. */
      ret = dfb_wm_add_window( stack, window );
      if (ret) {
-          if (window->surface) {
-               dfb_surface_detach_global( surface, &window->surface_reaction );
+          D_DERROR( ret, "Core/Windows: Failed to add window to manager!\n" );
+
+          if (window->surface)
                dfb_surface_unlink( &window->surface );
-          }
 
           if (window->primary_region)
                dfb_layer_region_unlink( &window->primary_region );
@@ -461,6 +483,8 @@ dfb_window_create( CoreWindowStack        *stack,
      /* Finally activate the object. */
      fusion_object_activate( &window->object );
 
+     fusion_reactor_direct( window->object.reactor, true );
+
      /* Unlock the window stack. */
      dfb_windowstack_unlock( stack );
 
@@ -475,6 +499,7 @@ dfb_window_destroy( CoreWindow *window )
 {
      DFBWindowEvent   evt;
      CoreWindowStack *stack;
+     BoundWindow     *bound, *next;
 
      D_ASSERT( window != NULL );
      D_ASSERT( DFB_WINDOW_INITIALIZED( window ) );
@@ -494,10 +519,24 @@ dfb_window_destroy( CoreWindow *window )
 
      /* Avoid multiple destructions. */
      if (DFB_WINDOW_DESTROYED( window )) {
+          D_DEBUG_AT( Core_Windows, "%p already destroyed.\n", window );
           dfb_windowstack_unlock( stack );
           return;
      }
 
+     /* Unbind bound windows. */
+     direct_list_foreach_safe (bound, next, window->bound_windows) {
+          direct_list_remove( &window->bound_windows, &bound->link );
+
+          bound->window->boundto = NULL;
+
+          SHFREE( stack->shmpool, bound );
+     }
+
+     /* Unbind this window. */
+     if (window->boundto)
+          dfb_window_unbind( window->boundto, window );
+
      /* Make sure the window is no longer visible. */
      dfb_window_set_opacity( window, 0 );
 
@@ -518,7 +557,6 @@ dfb_window_destroy( CoreWindow *window )
 
      /* Unlink the window's surface. */
      if (window->surface) {
-          dfb_surface_detach_global( window->surface, &window->surface_reaction );
           dfb_surface_unlink( &window->surface );
      }
 
@@ -551,6 +589,12 @@ dfb_window_change_stacking( CoreWindow             *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      config.stacking = stacking;
 
      /* Let the window manager do its work. */
@@ -577,6 +621,12 @@ dfb_window_raise( CoreWindow *window )
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      /* Let the window manager do its work. */
      ret = dfb_wm_restack_window( window, window, 1 );
 
@@ -601,6 +651,12 @@ dfb_window_lower( CoreWindow *window )
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      /* Let the window manager do its work. */
      ret = dfb_wm_restack_window( window, window, -1 );
 
@@ -625,6 +681,12 @@ dfb_window_raisetotop( CoreWindow *window )
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      /* Let the window manager do its work. */
      ret = dfb_wm_restack_window( window, NULL, 1 );
 
@@ -649,6 +711,12 @@ dfb_window_lowertobottom( CoreWindow *window )
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      /* Let the window manager do its work. */
      ret = dfb_wm_restack_window( window, NULL, 0 );
 
@@ -674,6 +742,12 @@ dfb_window_putatop( CoreWindow *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      /* Let the window manager do its work. */
      ret = dfb_wm_restack_window( window, lower, 1 );
 
@@ -699,6 +773,12 @@ dfb_window_putbelow( CoreWindow *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      /* Let the window manager do its work. */
      ret = dfb_wm_restack_window( window, upper, -1 );
 
@@ -709,34 +789,24 @@ dfb_window_putbelow( CoreWindow *window,
 }
 
 DFBResult
-dfb_window_move( CoreWindow *window,
-                 int         x,
-                 int         y,
-                 bool        relative )
+dfb_window_set_config( CoreWindow             *window,
+                       const CoreWindowConfig *config,
+                       CoreWindowConfigFlags   flags )
 {
      DFBResult         ret;
-     CoreWindowConfig  config;
      CoreWindowStack  *stack = window->stack;
 
      /* Lock the window stack. */
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
-     if (relative) {
-          config.bounds.x = window->config.bounds.x + x;
-          config.bounds.y = window->config.bounds.y + y;
-     }
-     else {
-          config.bounds.x = x;
-          config.bounds.y = y;
-     }
-
-     if (config.bounds.x == window->config.bounds.x && config.bounds.y == window->config.bounds.y) {
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
           dfb_windowstack_unlock( stack );
-          return DFB_OK;
+          return DFB_DESTROYED;
      }
 
-     ret = dfb_wm_set_window_config( window, &config, CWCF_POSITION );
+     ret = dfb_wm_set_window_config( window, config, flags );
 
      /* Unlock the window stack. */
      dfb_windowstack_unlock( stack );
@@ -744,6 +814,142 @@ dfb_window_move( CoreWindow *window,
      return ret;
 }
 
+static DFBResult
+move_window( CoreWindow *window,
+             int         x,
+             int         y )
+{
+     DFBResult         ret; 
+     CoreWindowConfig  config;
+     BoundWindow      *bound;
+
+     config.bounds.x = x;
+     config.bounds.y = y;
+
+     ret = dfb_wm_set_window_config( window, &config, CWCF_POSITION );
+     if (ret)
+          return ret;
+
+     direct_list_foreach (bound, window->bound_windows) {
+          move_window( bound->window,
+                       window->config.bounds.x + bound->x,
+                       window->config.bounds.y + bound->y );
+     }
+
+     return DFB_OK;
+}
+
+DFBResult
+dfb_window_move( CoreWindow *window,
+                 int         x,
+                 int         y,
+                 bool        relative )
+{
+     DFBResult        ret;    
+     CoreWindowStack *stack = window->stack;
+
+     /* Lock the window stack. */
+     if (dfb_windowstack_lock( stack ))
+          return DFB_FUSION;
+
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
+     if (window->boundto) {
+          dfb_windowstack_unlock( stack );
+          return DFB_UNSUPPORTED;
+     }
+
+     if (relative) {
+          x += window->config.bounds.x;
+          y += window->config.bounds.y;
+     }
+
+     if (x == window->config.bounds.x && y == window->config.bounds.y) {
+          dfb_windowstack_unlock( stack );
+          return DFB_OK;
+     }
+
+     ret = move_window( window, x, y );
+
+     /* Unlock the window stack. */
+     dfb_windowstack_unlock( stack );
+
+     return ret;
+}
+
+DFBResult
+dfb_window_set_bounds( CoreWindow *window,
+                       int         x,
+                       int         y,
+                       int         width,
+                       int         height )
+{
+     DFBResult         ret;
+     CoreWindowConfig  config;
+     CoreWindowStack  *stack = window->stack;
+     int               old_x;
+     int               old_y;
+
+     /* Lock the window stack. */
+     if (dfb_windowstack_lock( stack ))
+          return DFB_FUSION;
+
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
+     old_x = window->config.bounds.x;
+     old_y = window->config.bounds.y;
+
+     if (window->boundto) {
+          if (old_x != x || old_y != y) {
+               dfb_windowstack_unlock( stack );
+               return DFB_UNSUPPORTED;
+          }
+     }
+
+     config.bounds.x = x;
+     config.bounds.y = y;
+     config.bounds.w = width;
+     config.bounds.h = height;
+
+     if (window->config.bounds.x == x &&
+         window->config.bounds.y == y &&
+         window->config.bounds.w == width &&
+         window->config.bounds.h == height)
+     {
+          dfb_windowstack_unlock( stack );
+          return DFB_OK;
+     }
+
+     ret = dfb_wm_set_window_config( window, &config, CWCF_POSITION | CWCF_SIZE );
+     if (ret) {
+          dfb_windowstack_unlock( stack );
+          return ret;
+     }
+
+     if (old_x != x || old_y != y) {
+          BoundWindow *bound;
+
+          direct_list_foreach (bound, window->bound_windows) {
+               move_window( bound->window,
+                            window->config.bounds.x + bound->x,
+                            window->config.bounds.y + bound->y );
+          }
+     }
+
+     /* Unlock the window stack. */
+     dfb_windowstack_unlock( stack );
+
+     return DFB_OK;
+}
+
 DFBResult
 dfb_window_resize( CoreWindow   *window,
                    int           width,
@@ -766,6 +972,12 @@ dfb_window_resize( CoreWindow   *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      if (window->config.bounds.w == width && window->config.bounds.h == height) {
           dfb_windowstack_unlock( stack );
           return DFB_OK;
@@ -782,9 +994,115 @@ dfb_window_resize( CoreWindow   *window,
      return ret;
 }
 
+DFBResult
+dfb_window_bind( CoreWindow *window,
+                 CoreWindow *source,
+                 int         x,
+                 int         y )
+{
+     DFBResult        ret;
+     CoreWindowStack *stack = window->stack;
+     BoundWindow     *bound;
+
+     if (window == source)
+          return DFB_UNSUPPORTED;
+     
+     /* Lock the window stack. */
+     if (dfb_windowstack_lock( stack ))
+          return DFB_FUSION;
+
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
+     if (DFB_WINDOW_DESTROYED( source )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
+     bound = SHCALLOC( stack->shmpool, 1, sizeof(BoundWindow) );
+     if (!bound) {
+          dfb_windowstack_unlock( stack );
+          return DFB_NOSHAREDMEMORY;
+     }                    
+
+     if (source->boundto)
+          dfb_window_unbind( source->boundto, source );
+
+     ret = move_window( source,
+                        window->config.bounds.x + x,
+                        window->config.bounds.y + y );
+     if (ret) {
+          SHFREE( stack->shmpool, bound );
+          dfb_windowstack_unlock( stack );
+          return ret;
+     }
+
+     bound->window = source;
+     bound->x      = x;
+     bound->y      = y;
+
+     direct_list_append( &window->bound_windows, &bound->link );
+
+     source->boundto = window;
+
+     dfb_windowstack_unlock( stack );
+
+     return DFB_OK;
+}
+
+DFBResult
+dfb_window_unbind( CoreWindow *window,
+                   CoreWindow *source )
+{
+     CoreWindowStack *stack = window->stack;
+     BoundWindow     *bound;
+
+     /* Lock the window stack. */
+     if (dfb_windowstack_lock( stack ))
+          return DFB_FUSION;
+
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
+     if (DFB_WINDOW_DESTROYED( source )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
+     if (source->boundto != window) {
+          dfb_windowstack_unlock( stack );
+          return DFB_UNSUPPORTED;
+     }
+
+     direct_list_foreach (bound, window->bound_windows) {
+          if (bound->window == source) {
+               direct_list_remove( &window->bound_windows, &bound->link );
+
+               bound->window->boundto = NULL;
+
+               SHFREE( stack->shmpool, bound );
+
+               break;
+          }
+     }
+
+     if (!bound)
+          D_BUG( "window not found" );
+
+     dfb_windowstack_unlock( stack );
+
+     return bound ? DFB_OK : DFB_ITEMNOTFOUND;
+}
+     
 DFBResult
 dfb_window_set_colorkey( CoreWindow *window,
-                         __u32       color_key )
+                         u32         color_key )
 {
      DFBResult         ret;
      CoreWindowConfig  config;
@@ -794,6 +1112,12 @@ dfb_window_set_colorkey( CoreWindow *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      if (window->config.color_key == color_key) {
           dfb_windowstack_unlock( stack );
           return DFB_OK;
@@ -811,7 +1135,7 @@ dfb_window_set_colorkey( CoreWindow *window,
 
 DFBResult
 dfb_window_set_opacity( CoreWindow *window,
-                        __u8        opacity )
+                        u8          opacity )
 {
      DFBResult         ret;
      CoreWindowConfig  config;
@@ -821,6 +1145,12 @@ dfb_window_set_opacity( CoreWindow *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      if (window->config.opacity == opacity) {
           dfb_windowstack_unlock( stack );
           return DFB_OK;
@@ -854,6 +1184,12 @@ dfb_window_change_options( CoreWindow       *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      config.options = (window->config.options & ~disable) | enable;
 
      ret = dfb_wm_set_window_config( window, &config, CWCF_OPTIONS );
@@ -878,6 +1214,12 @@ dfb_window_set_opaque( CoreWindow      *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      config.opaque.x1 = 0;
      config.opaque.y1 = 0;
      config.opaque.x2 = window->config.bounds.w - 1;
@@ -912,6 +1254,12 @@ dfb_window_change_events( CoreWindow         *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      config.events = (window->config.events & ~disable) | enable;
 
      ret = dfb_wm_set_window_config( window, &config, CWCF_EVENTS );
@@ -923,82 +1271,70 @@ dfb_window_change_events( CoreWindow         *window,
 }
 
 DFBResult
-dfb_window_grab_keyboard( CoreWindow *window )
+dfb_window_set_key_selection( CoreWindow                    *window,
+                              DFBWindowKeySelection          selection,
+                              const DFBInputDeviceKeySymbol *keys,
+                              unsigned int                   num_keys )
 {
-     DFBResult        ret;
-     CoreWMGrab       grab;
-     CoreWindowStack *stack = window->stack;
+    DFBResult         ret;
+    CoreWindowConfig  config;
+    CoreWindowStack  *stack = window->stack;
 
-     /* Lock the window stack. */
-     if (dfb_windowstack_lock( stack ))
-          return DFB_FUSION;
+    D_ASSERT( selection == DWKS_ALL || selection == DWKS_NONE || selection == DWKS_LIST );
+    D_ASSERT( keys != NULL || selection != DWKS_LIST );
+    D_ASSERT( num_keys > 0 || selection != DWKS_LIST );
 
-     grab.target = CWMGT_KEYBOARD;
+    /* Lock the window stack. */
+    if (dfb_windowstack_lock( stack ))
+         return DFB_FUSION;
 
-     ret = dfb_wm_grab( window, &grab );
+    /* Never call WM after destroying the window. */
+    if (DFB_WINDOW_DESTROYED( window )) {
+         dfb_windowstack_unlock( stack );
+         return DFB_DESTROYED;
+    }
 
-     /* Unlock the window stack. */
-     dfb_windowstack_unlock( stack );
+    config.key_selection = selection;
+    config.keys          = (DFBInputDeviceKeySymbol*) keys; /* FIXME */
+    config.num_keys      = num_keys;
 
-     return ret;
+    ret = dfb_wm_set_window_config( window, &config, CWCF_KEY_SELECTION );
+
+    /* Unlock the window stack. */
+    dfb_windowstack_unlock( stack );
+
+    return ret;
 }
 
 DFBResult
-dfb_window_ungrab_keyboard( CoreWindow *window )
+dfb_window_change_grab( CoreWindow       *window,
+                        CoreWMGrabTarget  target,
+                        bool              grab )
 {
      DFBResult        ret;
-     CoreWMGrab       grab;
-     CoreWindowStack *stack = window->stack;
+     CoreWMGrab       wmgrab;
+     CoreWindowStack *stack;
+
+     D_ASSERT( window != NULL );
+
+     stack = window->stack;
 
      /* Lock the window stack. */
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
-     grab.target = CWMGT_KEYBOARD;
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
 
-     ret = dfb_wm_ungrab( window, &grab );
+     wmgrab.target = target;
 
-     /* Unlock the window stack. */
-     dfb_windowstack_unlock( stack );
-
-     return ret;
-}
-
-DFBResult
-dfb_window_grab_pointer( CoreWindow *window )
-{
-     DFBResult        ret;
-     CoreWMGrab       grab;
-     CoreWindowStack *stack = window->stack;
-
-     /* Lock the window stack. */
-     if (dfb_windowstack_lock( stack ))
-          return DFB_FUSION;
-
-     grab.target = CWMGT_POINTER;
-
-     ret = dfb_wm_grab( window, &grab );
-
-     /* Unlock the window stack. */
-     dfb_windowstack_unlock( stack );
-
-     return ret;
-}
-
-DFBResult
-dfb_window_ungrab_pointer( CoreWindow *window )
-{
-     DFBResult        ret;
-     CoreWMGrab       grab;
-     CoreWindowStack *stack = window->stack;
-
-     /* Lock the window stack. */
-     if (dfb_windowstack_lock( stack ))
-          return DFB_FUSION;
-
-     grab.target = CWMGT_POINTER;
-
-     ret = dfb_wm_ungrab( window, &grab );
+     if (grab)
+         ret = dfb_wm_grab( window, &wmgrab );
+     else
+         ret = dfb_wm_ungrab( window, &wmgrab );
 
      /* Unlock the window stack. */
      dfb_windowstack_unlock( stack );
@@ -1019,6 +1355,12 @@ dfb_window_grab_key( CoreWindow                 *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      grab.target    = CWMGT_KEY;
      grab.symbol    = symbol;
      grab.modifiers = modifiers;
@@ -1044,6 +1386,12 @@ dfb_window_ungrab_key( CoreWindow                 *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      grab.target    = CWMGT_KEY;
      grab.symbol    = symbol;
      grab.modifiers = modifiers;
@@ -1073,6 +1421,12 @@ dfb_window_repaint( CoreWindow          *window,
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      ret = dfb_wm_update_window( window, region, flags );
 
      /* Unlock the window stack. */
@@ -1081,6 +1435,62 @@ dfb_window_repaint( CoreWindow          *window,
      return ret;
 }
 
+const char *
+dfb_window_event_type_name( DFBWindowEventType type )
+{
+     switch (type) {
+          case DWET_POSITION:
+               return "POSITION";
+
+          case DWET_SIZE:
+               return "SIZE";
+
+          case DWET_CLOSE:
+               return "CLOSE";
+
+          case DWET_DESTROYED:
+               return "DESTROYED";
+
+          case DWET_GOTFOCUS:
+               return "GOTFOCUS";
+
+          case DWET_LOSTFOCUS:
+               return "LOSTFOCUS";
+
+          case DWET_KEYDOWN:
+               return "KEYDOWN";
+
+          case DWET_KEYUP:
+               return "KEYUP";
+
+          case DWET_BUTTONDOWN:
+               return "BUTTONDOWN";
+
+          case DWET_BUTTONUP:
+               return "BUTTONUP";
+
+          case DWET_MOTION:
+               return "MOTION";
+
+          case DWET_ENTER:
+               return "ENTER";
+
+          case DWET_LEAVE:
+               return "LEAVE";
+
+          case DWET_WHEEL:
+               return "WHEEL";
+
+          case DWET_POSITION_SIZE:
+               return "POSITION_SIZE";
+
+          default:
+               break;
+     }
+
+     return "<invalid>";
+}
+
 void
 dfb_window_post_event( CoreWindow     *window,
                        DFBWindowEvent *event )
@@ -1116,6 +1526,8 @@ dfb_window_send_configuration( CoreWindow *window )
 {
      DFBWindowEvent event;
 
+     D_ASSUME( !DFB_WINDOW_DESTROYED( window ) );
+
      event.type = DWET_POSITION_SIZE;
      event.x    = window->config.bounds.x;
      event.y    = window->config.bounds.y;
@@ -1137,6 +1549,12 @@ dfb_window_request_focus( CoreWindow *window )
      if (dfb_windowstack_lock( stack ))
           return DFB_FUSION;
 
+     /* Never call WM after destroying the window. */
+     if (DFB_WINDOW_DESTROYED( window )) {
+          dfb_windowstack_unlock( stack );
+          return DFB_DESTROYED;
+     }
+
      ret = dfb_wm_request_focus( window );
 
      /* Unlock the window stack. */
@@ -1163,42 +1581,6 @@ dfb_window_surface( const CoreWindow *window )
 
 /******************************************************************************/
 
-/*
- * listen to the window's surface
- */
-ReactionResult
-_dfb_window_surface_listener( const void *msg_data, void *ctx )
-{
-     const CoreSurfaceNotification *notification = msg_data;
-     CoreWindow                    *window       = ctx;
-
-     (void) window;
-
-     D_ASSERT( notification != NULL );
-     D_ASSERT( notification->surface != NULL );
-
-     D_ASSERT( window != NULL );
-     D_ASSERT( window->stack != NULL );
-     D_ASSERT( window->surface == notification->surface );
-
-     if (notification->flags & CSNF_DESTROY) {
-          D_WARN( "window surface destroyed" );
-          return RS_REMOVE;
-     }
-
-     if (notification->flags & (CSNF_PALETTE_CHANGE | CSNF_PALETTE_UPDATE)) {
-/*          if (window->window_data) {
-               CoreLayer *layer = dfb_layer_at( window->stack->layer_id );
-
-               D_ASSERT( layer != NULL );
-
-               dfb_layer_update_window( layer, window, CWUF_PALETTE );
-          }*/
-     }
-
-     return RS_OK;
-}
-
 static bool
 core_window_filter( CoreWindow *window, const DFBWindowEvent *event )
 {
