@@ -1,5 +1,5 @@
 /*
-   (c) Copyright 2001-2007  The DirectFB Organization (directfb.org)
+   (c) Copyright 2001-2008  The world wide DirectFB Open Source Community (directfb.org)
    (c) Copyright 2000-2004  Convergence (integrated media) GmbH
 
    All rights reserved.
@@ -47,8 +47,8 @@
 #include <core/fonts.h>
 #include <core/state.h>
 #include <core/palette.h>
-#include <core/surfaces.h>
-#include <core/surfacemanager.h>
+#include <core/surface.h>
+#include <core/surface_buffer.h>
 
 #include <media/idirectfbfont.h>
 
@@ -57,6 +57,7 @@
 
 #include <misc/util.h>
 
+#include <direct/debug.h>
 #include <direct/interface.h>
 #include <direct/mem.h>
 #include <direct/memcpy.h>
@@ -79,6 +80,7 @@ void
 IDirectFBSurface_Destruct( IDirectFBSurface *thiz )
 {
      IDirectFBSurface_data *data;
+     IDirectFBSurface      *parent;
 
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
 
@@ -86,11 +88,33 @@ IDirectFBSurface_Destruct( IDirectFBSurface *thiz )
 
      data = thiz->priv;
 
+     D_ASSERT( data != NULL );
+     D_ASSERT( data->children_data == NULL );
+
+     parent = data->parent;
+     if (parent) {
+          IDirectFBSurface_data *parent_data;
+
+          D_MAGIC_ASSERT( (IAny*) parent, DirectInterface );
+
+          parent_data = parent->priv;
+          D_ASSERT( parent_data != NULL );
+
+          pthread_mutex_lock( &parent_data->children_lock );
+
+          direct_list_remove( &parent_data->children_data, &data->link );
+
+          pthread_mutex_unlock( &parent_data->children_lock );
+     }
+
      if (data->surface)
           dfb_surface_detach( data->surface, &data->reaction );
 
+     dfb_state_stop_drawing( &data->state );
+
      dfb_state_set_destination( &data->state, NULL );
      dfb_state_set_source( &data->state, NULL );
+     dfb_state_set_source_mask( &data->state, NULL );
 
      dfb_state_destroy( &data->state );
 
@@ -110,15 +134,20 @@ IDirectFBSurface_Destruct( IDirectFBSurface *thiz )
 
      if (data->surface) {
           if (data->locked)
-               dfb_surface_unlock( data->surface, data->locked - 1 );
+               dfb_surface_unlock_buffer( data->surface, &data->lock );
 
           dfb_surface_unref( data->surface );
      }
 
+     pthread_mutex_destroy( &data->children_lock );
+
      DIRECT_DEALLOCATE_INTERFACE( thiz );
+
+     if (parent)
+          parent->Release( parent );
 }
 
-static DFBResult
+static DirectResult
 IDirectFBSurface_AddRef( IDirectFBSurface *thiz )
 {
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
@@ -130,7 +159,7 @@ IDirectFBSurface_AddRef( IDirectFBSurface *thiz )
      return DFB_OK;
 }
 
-static DFBResult
+static DirectResult
 IDirectFBSurface_Release( IDirectFBSurface *thiz )
 {
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
@@ -158,7 +187,7 @@ IDirectFBSurface_GetPixelFormat( IDirectFBSurface      *thiz,
      if (!format)
           return DFB_INVARG;
 
-     *format = data->surface->format;
+     *format = data->surface->config.format;
 
      return DFB_OK;
 }
@@ -331,7 +360,7 @@ IDirectFBSurface_SetPalette( IDirectFBSurface *thiz,
      if (!palette)
           return DFB_INVARG;
 
-     if (! DFB_PIXELFORMAT_IS_INDEXED( surface->format ))
+     if (! DFB_PIXELFORMAT_IS_INDEXED( surface->config.format ))
           return DFB_UNSUPPORTED;
 
      palette_data = (IDirectFBPalette_data*) palette->priv;
@@ -370,10 +399,9 @@ IDirectFBSurface_Lock( IDirectFBSurface *thiz,
                        DFBSurfaceLockFlags flags,
                        void **ret_ptr, int *ret_pitch )
 {
-     DFBResult  ret;
-     int        front;
-     int        pitch;
-     void      *ptr;
+     DFBResult              ret;
+     CoreSurfaceBufferRole  role   = CSBR_FRONT;
+     CoreSurfaceAccessFlags access = CSAF_NONE;
 
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
@@ -391,19 +419,23 @@ IDirectFBSurface_Lock( IDirectFBSurface *thiz,
      if (!data->area.current.w || !data->area.current.h)
           return DFB_INVAREA;
 
-     front = (flags & DSLF_WRITE) ? 0 : 1;
+     if (flags & DSLF_READ)
+          access |= CSAF_CPU_READ;
 
-     ret = dfb_surface_soft_lock( data->core, data->surface, flags, &ptr, &pitch, front );
+     if (flags & DSLF_WRITE) {
+          access |= CSAF_CPU_WRITE;
+          role = CSBR_BACK;
+     }
+
+     ret = dfb_surface_lock_buffer( data->surface, role, access, &data->lock );
      if (ret)
           return ret;
 
-     ptr += pitch * data->area.current.y +
-            DFB_BYTES_PER_LINE( data->surface->format, data->area.current.x );
+     data->locked = true;
 
-     data->locked = front + 1;
-
-     *ret_ptr   = ptr;
-     *ret_pitch = pitch;
+     *ret_ptr   = data->lock.addr + data->lock.pitch * data->area.current.y +
+                  DFB_BYTES_PER_LINE( data->surface->config.format, data->area.current.x );
+     *ret_pitch = data->lock.pitch;
 
      return DFB_OK;
 }
@@ -412,9 +444,6 @@ static DFBResult
 IDirectFBSurface_GetFramebufferOffset( IDirectFBSurface *thiz,
                                        int *offset )
 {
-     int            front;
-     SurfaceBuffer *buffer;
-
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
      if (!data->surface)
@@ -426,22 +455,13 @@ IDirectFBSurface_GetFramebufferOffset( IDirectFBSurface *thiz,
      if (!data->locked)
           return DFB_ACCESSDENIED;
 
-     front = data->locked - 1; /* As set by Lock(). */
-
-     if (front)
-          buffer = data->surface->front_buffer;
-     else
-          buffer = data->surface->back_buffer;
-
-     if (!buffer)
-          return DFB_BUG;
-
-     if (!buffer->video.locked) {
-          /* The surface is probably in a system buffer if video is unlocked. */
+     if (!data->lock.phys) {
+          /* The surface is probably in a system buffer if there's no physical address. */
           return DFB_UNSUPPORTED;
      }
 
-     *offset = buffer->video.offset;
+     *offset = data->lock.offset;
+
      return DFB_OK;
 }
 
@@ -452,10 +472,14 @@ IDirectFBSurface_Unlock( IDirectFBSurface *thiz )
 
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
 
-     if (data->locked)
-          dfb_surface_unlock( data->surface, data->locked - 1 );
+     if (!data->surface)
+          return DFB_DESTROYED;
 
-     data->locked = 0;
+     if (data->locked) {
+          dfb_surface_unlock_buffer( data->surface, &data->lock );
+
+          data->locked = false;
+     }
 
      return DFB_OK;
 }
@@ -465,6 +489,7 @@ IDirectFBSurface_Flip( IDirectFBSurface    *thiz,
                        const DFBRegion     *region,
                        DFBSurfaceFlipFlags  flags )
 {
+     DFBResult    ret;
      DFBRegion    reg;
      CoreSurface *surface;
 
@@ -479,13 +504,26 @@ IDirectFBSurface_Flip( IDirectFBSurface    *thiz,
      if (data->locked)
           return DFB_LOCKED;
 
-     if (!(surface->caps & DSCAPS_FLIPPING))
+     if (!(surface->config.caps & DSCAPS_FLIPPING))
           return DFB_UNSUPPORTED;
 
      if (!data->area.current.w || !data->area.current.h ||
          (region && (region->x1 > region->x2 || region->y1 > region->y2)))
           return DFB_INVAREA;
 
+     IDirectFBSurface_StopAll( data );
+
+     /* FIXME: This is a temporary workaround for LiTE. */
+     if (data->parent) {
+          IDirectFBSurface_data *parent_data;
+
+          DIRECT_INTERFACE_GET_DATA_FROM( data->parent, parent_data, IDirectFBSurface );
+
+          /* Signal end of sequence of operations. */
+          dfb_state_lock( &parent_data->state );
+          dfb_state_stop_drawing( &parent_data->state );
+          dfb_state_unlock( &parent_data->state );
+     }
 
      dfb_region_from_rectangle( &reg, &data->area.current );
 
@@ -498,11 +536,19 @@ IDirectFBSurface_Flip( IDirectFBSurface    *thiz,
                return DFB_INVAREA;
      }
 
-     D_DEBUG_AT( Surface, "  -> %d, %d - %dx%d\n", DFB_RECTANGLE_VALS_FROM_REGION( &reg ) );
+     D_DEBUG_AT( Surface, "  ->      %4d,%4d-%4dx%4d\n", DFB_RECTANGLE_VALS_FROM_REGION( &reg ) );
 
      if (!(flags & DSFLIP_BLIT) && reg.x1 == 0 && reg.y1 == 0 &&
-         reg.x2 == surface->width - 1 && reg.y2 == surface->height - 1)
-          dfb_surface_flip_buffers( data->surface, false );
+         reg.x2 == surface->config.size.w - 1 && reg.y2 == surface->config.size.h - 1)
+     {
+          ret = dfb_surface_lock( data->surface );
+          if (ret)
+               return ret;
+
+          dfb_surface_flip( data->surface, false );
+
+          dfb_surface_unlock( data->surface );
+     }
      else
           dfb_back_to_front_copy( data->surface, &reg );
 
@@ -520,7 +566,7 @@ IDirectFBSurface_SetField( IDirectFBSurface    *thiz,
      if (!data->surface)
           return DFB_DESTROYED;
 
-     if (!(data->surface->caps & DSCAPS_INTERLACED))
+     if (!(data->surface->config.caps & DSCAPS_INTERLACED))
           return DFB_UNSUPPORTED;
 
      if (field < 0 || field > 1)
@@ -538,12 +584,13 @@ IDirectFBSurface_Clear( IDirectFBSurface *thiz,
      DFBColor                old_color;
      unsigned int            old_index;
      DFBSurfaceDrawingFlags  old_flags;
+     DFBSurfaceRenderOptions old_options;
      CoreSurface            *surface;
      DFBColor                color = { a, r, g, b };
 
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, 0x%08x )\n", __FUNCTION__, thiz, PIXEL_ARGB(a,r,g,b) );
 
      surface = data->surface;
      if (!surface)
@@ -556,15 +603,19 @@ IDirectFBSurface_Clear( IDirectFBSurface *thiz,
           return DFB_LOCKED;
 
      /* save current color and drawing flags */
-     old_color = data->state.color;
-     old_index = data->state.color_index;
-     old_flags = data->state.drawingflags;
+     old_color   = data->state.color;
+     old_index   = data->state.color_index;
+     old_flags   = data->state.drawingflags;
+     old_options = data->state.render_options;
 
      /* set drawing flags */
      dfb_state_set_drawing_flags( &data->state, DSDRAW_NOFX );
 
+     /* set render options */
+     dfb_state_set_render_options( &data->state, DSRO_NONE );
+
      /* set color */
-     if (DFB_PIXELFORMAT_IS_INDEXED( surface->format ))
+     if (DFB_PIXELFORMAT_IS_INDEXED( surface->config.format ))
           dfb_state_set_color_index( &data->state,
                                      dfb_palette_search( surface->palette, r, g, b, a ) );
 
@@ -580,8 +631,11 @@ IDirectFBSurface_Clear( IDirectFBSurface *thiz,
      /* restore drawing flags */
      dfb_state_set_drawing_flags( &data->state, old_flags );
 
+     /* restore render options */
+     dfb_state_set_render_options( &data->state, old_options );
+
      /* restore color */
-     if (DFB_PIXELFORMAT_IS_INDEXED( surface->format ))
+     if (DFB_PIXELFORMAT_IS_INDEXED( surface->config.format ))
           dfb_state_set_color_index( &data->state, old_index );
 
      dfb_state_set_color( &data->state, &old_color );
@@ -596,7 +650,7 @@ IDirectFBSurface_SetClip( IDirectFBSurface *thiz, const DFBRegion *clip )
 
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, %p )\n", __FUNCTION__, thiz, clip );
 
 
      if (!data->area.current.w || !data->area.current.h)
@@ -605,10 +659,14 @@ IDirectFBSurface_SetClip( IDirectFBSurface *thiz, const DFBRegion *clip )
      if (clip) {
           newclip = DFB_REGION_INIT_TRANSLATED( clip, data->area.wanted.x, data->area.wanted.y );
 
+          D_DEBUG_AT( Surface, "  <-      %4d,%4d-%4dx%4d\n", DFB_RECTANGLE_VALS_FROM_REGION(&newclip) );
+
           if (!dfb_unsafe_region_rectangle_intersect( &newclip,
                                                       &data->area.wanted ))
                return DFB_INVARG;
 
+          D_DEBUG_AT( Surface, "  ->      %4d,%4d-%4dx%4d\n", DFB_RECTANGLE_VALS_FROM_REGION(&newclip) );
+
           data->clip_set = true;
           data->clip_wanted = newclip;
 
@@ -620,6 +678,8 @@ IDirectFBSurface_SetClip( IDirectFBSurface *thiz, const DFBRegion *clip )
           data->clip_set = false;
      }
 
+     D_DEBUG_AT( Surface, "  => CLIP %4d,%4d-%4dx%4d\n", DFB_RECTANGLE_VALS_FROM_REGION(&newclip) );
+
      dfb_state_set_clip( &data->state, &newclip );
 
      return DFB_OK;
@@ -640,6 +700,8 @@ IDirectFBSurface_GetClip( IDirectFBSurface *thiz, DFBRegion *ret_clip )
 
      *ret_clip = DFB_REGION_INIT_TRANSLATED( &data->state.clip, -data->area.wanted.x, -data->area.wanted.y );
 
+     D_DEBUG_AT( Surface, "  ->      %4d,%4d-%4dx%4d\n", DFB_RECTANGLE_VALS_FROM_REGION(ret_clip) );
+
      return DFB_OK;
 }
 
@@ -652,7 +714,7 @@ IDirectFBSurface_SetColor( IDirectFBSurface *thiz,
 
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, COLOR 0x%08x )\n", __FUNCTION__, thiz, PIXEL_ARGB(a, r, g, b) );
 
      surface = data->surface;
      if (!surface)
@@ -660,7 +722,7 @@ IDirectFBSurface_SetColor( IDirectFBSurface *thiz,
 
      dfb_state_set_color( &data->state, &color );
 
-     if (DFB_PIXELFORMAT_IS_INDEXED( surface->format ))
+     if (DFB_PIXELFORMAT_IS_INDEXED( surface->config.format ))
           dfb_state_set_color_index( &data->state,
                                      dfb_palette_search( surface->palette, r, g, b, a ) );
 
@@ -676,13 +738,13 @@ IDirectFBSurface_SetColorIndex( IDirectFBSurface *thiz,
 
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, COLOR INDEX %3u )\n", __FUNCTION__, thiz, index );
 
      surface = data->surface;
      if (!surface)
           return DFB_DESTROYED;
 
-     if (! DFB_PIXELFORMAT_IS_INDEXED( surface->format ))
+     if (! DFB_PIXELFORMAT_IS_INDEXED( surface->config.format ))
           return DFB_UNSUPPORTED;
 
      palette = surface->palette;
@@ -705,7 +767,7 @@ IDirectFBSurface_SetSrcBlendFunction( IDirectFBSurface        *thiz,
 {
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, %d )\n", __FUNCTION__, thiz, src );
 
      switch (src) {
           case DSBF_ZERO:
@@ -721,6 +783,9 @@ IDirectFBSurface_SetSrcBlendFunction( IDirectFBSurface        *thiz,
           case DSBF_SRCALPHASAT:
                dfb_state_set_src_blend( &data->state, src );
                return DFB_OK;
+
+          default:
+               break;
      }
 
      return DFB_INVARG;
@@ -732,7 +797,7 @@ IDirectFBSurface_SetDstBlendFunction( IDirectFBSurface        *thiz,
 {
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, %d )\n", __FUNCTION__, thiz, dst );
 
      switch (dst) {
           case DSBF_ZERO:
@@ -748,6 +813,9 @@ IDirectFBSurface_SetDstBlendFunction( IDirectFBSurface        *thiz,
           case DSBF_SRCALPHASAT:
                dfb_state_set_dst_blend( &data->state, dst );
                return DFB_OK;
+
+          default:
+               break;
      }
 
      return DFB_INVARG;
@@ -762,7 +830,7 @@ IDirectFBSurface_SetPorterDuff( IDirectFBSurface         *thiz,
 
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, %d )\n", __FUNCTION__, thiz, rule );
 
 
      switch (rule) {
@@ -848,11 +916,11 @@ IDirectFBSurface_SetSrcColorKey( IDirectFBSurface *thiz,
      data->src_key.g = g;
      data->src_key.b = b;
 
-     if (DFB_PIXELFORMAT_IS_INDEXED( surface->format ))
+     if (DFB_PIXELFORMAT_IS_INDEXED( surface->config.format ))
           data->src_key.value = dfb_palette_search( surface->palette,
                                                     r, g, b, 0x80 );
      else
-          data->src_key.value = dfb_color_to_pixel( surface->format, r, g, b );
+          data->src_key.value = dfb_color_to_pixel( surface->config.format, r, g, b );
 
      /* The new key won't be applied to this surface's state.
         The key will be taken by the destination surface to apply it
@@ -876,7 +944,7 @@ IDirectFBSurface_SetSrcColorKeyIndex( IDirectFBSurface *thiz,
      if (!surface)
           return DFB_DESTROYED;
 
-     if (! DFB_PIXELFORMAT_IS_INDEXED( surface->format ))
+     if (! DFB_PIXELFORMAT_IS_INDEXED( surface->config.format ))
           return DFB_UNSUPPORTED;
 
      palette = surface->palette;
@@ -919,11 +987,11 @@ IDirectFBSurface_SetDstColorKey( IDirectFBSurface *thiz,
      data->dst_key.g = g;
      data->dst_key.b = b;
 
-     if (DFB_PIXELFORMAT_IS_INDEXED( surface->format ))
+     if (DFB_PIXELFORMAT_IS_INDEXED( surface->config.format ))
           data->dst_key.value = dfb_palette_search( surface->palette,
                                                     r, g, b, 0x80 );
      else
-          data->dst_key.value = dfb_color_to_pixel( surface->format, r, g, b );
+          data->dst_key.value = dfb_color_to_pixel( surface->config.format, r, g, b );
 
      dfb_state_set_dst_colorkey( &data->state, data->dst_key.value );
 
@@ -945,7 +1013,7 @@ IDirectFBSurface_SetDstColorKeyIndex( IDirectFBSurface *thiz,
      if (!surface)
           return DFB_DESTROYED;
 
-     if (! DFB_PIXELFORMAT_IS_INDEXED( surface->format ))
+     if (! DFB_PIXELFORMAT_IS_INDEXED( surface->config.format ))
           return DFB_UNSUPPORTED;
 
      palette = surface->palette;
@@ -979,7 +1047,7 @@ IDirectFBSurface_SetIndexTranslation( IDirectFBSurface *thiz,
      if (!surface)
           return DFB_DESTROYED;
 
-     if (! DFB_PIXELFORMAT_IS_INDEXED( surface->format ))
+     if (! DFB_PIXELFORMAT_IS_INDEXED( surface->config.format ))
           return DFB_UNSUPPORTED;
 
      if (!indices && num_indices > 0)
@@ -999,7 +1067,7 @@ IDirectFBSurface_SetFont( IDirectFBSurface *thiz,
 
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, %p )\n", __FUNCTION__, thiz, font );
 
      if (data->font != font) {
          if (font) {
@@ -1066,7 +1134,7 @@ IDirectFBSurface_SetDrawingFlags( IDirectFBSurface       *thiz,
 {
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, 0x%08x )\n", __FUNCTION__, thiz, flags );
 
      dfb_state_set_drawing_flags( &data->state, flags );
 
@@ -1082,6 +1150,7 @@ IDirectFBSurface_FillRectangle( IDirectFBSurface *thiz,
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "  -> [%2d] %4d,%4d-%4dx%4d\n", 0, x, y, w, h );
 
      if (!data->surface)
           return DFB_DESTROYED;
@@ -1112,6 +1181,7 @@ IDirectFBSurface_DrawLine( IDirectFBSurface *thiz,
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "  -> [%2d] %4d,%4d-%4d,%4d\n", 0, x1, y1, x2, y2 );
 
      if (!data->surface)
           return DFB_DESTROYED;
@@ -1123,7 +1193,7 @@ IDirectFBSurface_DrawLine( IDirectFBSurface *thiz,
      if (data->locked)
           return DFB_LOCKED;
 
-     if (x1 == x2 || y1 == y2) {
+     if ((x1 == x2 || y1 == y2) && !(data->state.render_options & DSRO_MATRIX)) {
           DFBRectangle rect;
 
           if (x1 <= x2) {
@@ -1168,11 +1238,11 @@ IDirectFBSurface_DrawLines( IDirectFBSurface *thiz,
                             const DFBRegion  *lines,
                             unsigned int      num_lines )
 {
-     DFBRegion *local_lines = alloca(sizeof(DFBRegion) * num_lines);
+     unsigned int i;
 
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, %p [%d] )\n", __FUNCTION__, thiz, lines, num_lines );
 
      if (!data->surface)
           return DFB_DESTROYED;
@@ -1187,21 +1257,53 @@ IDirectFBSurface_DrawLines( IDirectFBSurface *thiz,
      if (!lines || !num_lines)
           return DFB_INVARG;
 
-     if (data->area.wanted.x || data->area.wanted.y) {
-          unsigned int i;
-
-          for (i=0; i<num_lines; i++) {
-               local_lines[i].x1 = lines[i].x1 + data->area.wanted.x;
-               local_lines[i].x2 = lines[i].x2 + data->area.wanted.x;
-               local_lines[i].y1 = lines[i].y1 + data->area.wanted.y;
-               local_lines[i].y2 = lines[i].y2 + data->area.wanted.y;
-          }
+     /* Check if all lines are either horizontal or vertical */
+     for (i=0; i<num_lines; i++) {
+          if (lines[i].x1 != lines[i].x2 && lines[i].y1 != lines[i].y2)
+               break;
      }
-     else
-          /* clipping may modify lines, so we copy them */
-          direct_memcpy( local_lines, lines, sizeof(DFBRegion) * num_lines );
 
-     dfb_gfxcard_drawlines( local_lines, num_lines, &data->state );
+     /* Use real line drawing? */
+     if (i < num_lines) {
+          DFBRegion *local_lines = alloca(sizeof(DFBRegion) * num_lines);
+          
+          if (data->area.wanted.x || data->area.wanted.y) {
+               for (i=0; i<num_lines; i++) {
+                    local_lines[i].x1 = lines[i].x1 + data->area.wanted.x;
+                    local_lines[i].x2 = lines[i].x2 + data->area.wanted.x;
+                    local_lines[i].y1 = lines[i].y1 + data->area.wanted.y;
+                    local_lines[i].y2 = lines[i].y2 + data->area.wanted.y;
+               }
+          }
+          else
+               /* clipping may modify lines, so we copy them */
+               direct_memcpy( local_lines, lines, sizeof(DFBRegion) * num_lines );
+
+          dfb_gfxcard_drawlines( local_lines, num_lines, &data->state );
+     }
+     /* Optimized rectangle drawing */
+     else {
+          DFBRectangle *local_rects = alloca(sizeof(DFBRectangle) * num_lines);
+          
+          for (i=0; i<num_lines; i++) {
+               /* Vertical line? */
+               if (lines[i].x1 == lines[i].x2) {
+                    local_rects[i].x = data->area.wanted.x + lines[i].x1;
+                    local_rects[i].y = data->area.wanted.y + MIN( lines[i].y1, lines[i].y2 );
+                    local_rects[i].w = 1;
+                    local_rects[i].h = ABS( lines[i].y2 - lines[i].y1 ) + 1;
+               }
+               /* Horizontal line */
+               else {
+                    local_rects[i].x = data->area.wanted.x + MIN( lines[i].x1, lines[i].x2 );
+                    local_rects[i].y = data->area.wanted.y + lines[i].y1;
+                    local_rects[i].w = ABS( lines[i].x2 - lines[i].x1 ) + 1;
+                    local_rects[i].h = 1;
+               }
+          }
+
+          dfb_gfxcard_fillrectangles( local_rects, num_lines, &data->state );
+     }
 
      return DFB_OK;
 }
@@ -1215,6 +1317,7 @@ IDirectFBSurface_DrawRectangle( IDirectFBSurface *thiz,
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "  -> [%2d] %4d,%4d-%4dx%4d\n", 0, x, y, w, h );
 
      if (!data->surface)
           return DFB_DESTROYED;
@@ -1248,6 +1351,7 @@ IDirectFBSurface_FillTriangle( IDirectFBSurface *thiz,
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "  -> [%2d] %4d,%4d-%4d,%4d-%4d,%4d\n", 0, x1, y1, x2, y2, x3, y3 );
 
      if (!data->surface)
           return DFB_DESTROYED;
@@ -1266,7 +1370,7 @@ IDirectFBSurface_FillTriangle( IDirectFBSurface *thiz,
      tri.x3 += data->area.wanted.x;
      tri.y3 += data->area.wanted.y;
 
-     dfb_gfxcard_filltriangle( &tri, &data->state );
+     dfb_gfxcard_filltriangles( &tri, 1, &data->state );
 
      return DFB_OK;
 }
@@ -1278,7 +1382,10 @@ IDirectFBSurface_FillRectangles( IDirectFBSurface   *thiz,
 {
      DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
 
-     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+     D_DEBUG_AT( Surface, "%s( %p, %p [%d] )\n", __FUNCTION__, thiz, rects, num_rects );
+
+     DFB_RECTANGLES_DEBUG_AT( Surface, rects, num_rects );
+
 
      if (!data->surface)
           return DFB_DESTROYED;
@@ -1363,6 +1470,58 @@ IDirectFBSurface_FillSpans( IDirectFBSurface *thiz,
      return DFB_OK;
 }
 
+static DFBResult
+IDirectFBSurface_FillTriangles( IDirectFBSurface  *thiz,
+                                const DFBTriangle *tris,
+                                unsigned int       num_tris )
+{
+     DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
+
+     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+
+     if (!data->surface)
+          return DFB_DESTROYED;
+
+
+     if (!data->area.current.w || !data->area.current.h)
+          return DFB_INVAREA;
+
+     if (data->locked)
+          return DFB_LOCKED;
+
+     if (!tris || !num_tris)
+          return DFB_INVARG;
+
+     if (data->area.wanted.x || data->area.wanted.y) {
+          unsigned int  i;
+          DFBTriangle  *local_tris;
+          bool          malloced = (num_tris > 170);
+
+          if (malloced)
+               local_tris = D_MALLOC( sizeof(DFBTriangle) * num_tris );
+          else
+               local_tris = alloca( sizeof(DFBTriangle) * num_tris );
+
+          for (i=0; i<num_tris; i++) {
+               local_tris[i].x1 = tris[i].x1 + data->area.wanted.x;
+               local_tris[i].y1 = tris[i].y1 + data->area.wanted.y;
+               local_tris[i].x2 = tris[i].x2 + data->area.wanted.x;
+               local_tris[i].y2 = tris[i].y2 + data->area.wanted.y;
+               local_tris[i].x3 = tris[i].x3 + data->area.wanted.x;
+               local_tris[i].y3 = tris[i].y3 + data->area.wanted.y;
+          }
+
+          dfb_gfxcard_filltriangles( local_tris, num_tris, &data->state );
+
+          if (malloced)
+               D_FREE( local_tris );
+     }
+     else
+          dfb_gfxcard_filltriangles( tris, num_tris, &data->state );
+
+     return DFB_OK;
+}
+
 static DFBResult
 IDirectFBSurface_SetBlittingFlags( IDirectFBSurface        *thiz,
                                    DFBSurfaceBlittingFlags  flags )
@@ -1389,6 +1548,9 @@ IDirectFBSurface_Blit( IDirectFBSurface   *thiz,
 
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
 
+     if (sr)
+          D_DEBUG_AT( Surface, "  -> [%2d] %4d,%4d-%4dx%4d <- %4d,%4d\n", 0, dx, dy, sr->w, sr->h, sr->x, sr->y );
+
      if (!data->surface)
           return DFB_DESTROYED;
 
@@ -1455,6 +1617,9 @@ IDirectFBSurface_TileBlit( IDirectFBSurface   *thiz,
 
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
 
+     if (sr)
+          D_DEBUG_AT( Surface, "  -> [%2d] %4d,%4d-%4dx%4d <- %4d,%4d\n", 0, dx, dy, sr->w, sr->h, sr->x, sr->y );
+
      if (!data->surface)
           return DFB_DESTROYED;
 
@@ -1755,8 +1920,8 @@ IDirectFBSurface_TextureTriangles( IDirectFBSurface     *thiz,
 
      /* TODO: pass indices through to driver */
      if (src_sub) {
-          float oowidth  = 1.0f / src_data->surface->width;
-          float ooheight = 1.0f / src_data->surface->height;
+          float oowidth  = 1.0f / src_data->surface->config.size.w;
+          float ooheight = 1.0f / src_data->surface->config.size.h;
 
           float s0 = src_data->area.wanted.x * oowidth;
           float t0 = src_data->area.wanted.y * ooheight;
@@ -2036,14 +2201,14 @@ IDirectFBSurface_GetSubSurface( IDirectFBSurface    *thiz,
           dfb_rectangle_intersect( &granted, &data->area.granted );
           
           /* Construct */
-          ret = IDirectFBSurface_Construct( *surface, 
+          ret = IDirectFBSurface_Construct( *surface, thiz,
                                             &wanted, &granted, &data->area.insets,
                                             data->surface,
                                             data->caps | DSCAPS_SUBSURFACE, data->core );
      }
      else {
           /* Construct */
-          ret = IDirectFBSurface_Construct( *surface, 
+          ret = IDirectFBSurface_Construct( *surface, thiz,
                                             NULL, NULL, &data->area.insets,
                                             data->surface, 
                                             data->caps | DSCAPS_SUBSURFACE, data->core );
@@ -2052,6 +2217,89 @@ IDirectFBSurface_GetSubSurface( IDirectFBSurface    *thiz,
      return ret;
 }
 
+static DFBResult
+IDirectFBSurface_MakeSubSurface( IDirectFBSurface   *thiz,
+                                 IDirectFBSurface   *from,
+                                 const DFBRectangle *rect )
+{
+     CoreSurface           *surface;
+     DFBRectangle           wanted, granted;
+     DFBRectangle           full_rect;
+     IDirectFBSurface_data *from_data;
+
+     DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
+
+     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+
+     /* Check arguments */
+     if (!from)
+          return DFB_INVARG;
+
+     surface = data->surface;
+     if (!surface)
+          return DFB_DESTROYED;
+
+     DIRECT_INTERFACE_GET_DATA_FROM(from, from_data, IDirectFBSurface);
+
+     /* Check if CoreSurface is the same */
+     if (from_data->surface != surface)
+          return DFB_UNSUPPORTED;
+
+
+     full_rect.x = 0;
+     full_rect.y = 0;
+     full_rect.w = surface->config.size.w;
+     full_rect.h = surface->config.size.h;
+
+     if (rect || from_data->limit_set) {
+          /* Compute wanted rectangle */
+          if (rect) {
+               wanted = *rect;
+
+               wanted.x += from_data->area.wanted.x;
+               wanted.y += from_data->area.wanted.y;
+
+               if (wanted.w <= 0 || wanted.h <= 0) {
+                    wanted.w = 0;
+                    wanted.h = 0;
+               }
+          }
+          else {
+               wanted = from_data->area.wanted;
+          }
+          
+          /* Compute granted rectangle */
+          granted = wanted;
+
+          dfb_rectangle_intersect( &granted, &from_data->area.granted );
+     }
+     else {
+          wanted  = full_rect;
+          granted = full_rect;
+     }
+
+
+     data->caps |= DSCAPS_SUBSURFACE;
+
+
+     data->area.wanted  = wanted;
+     data->area.granted = granted;
+
+     data->area.current = data->area.granted;
+     dfb_rectangle_intersect( &data->area.current, &full_rect );
+
+
+     data->state.clip.x1   = data->area.current.x;
+     data->state.clip.y1   = data->area.current.y;
+     data->state.clip.x2   = data->area.current.x + (data->area.current.w ? : 1) - 1;
+     data->state.clip.y2   = data->area.current.y + (data->area.current.h ? : 1) - 1;
+
+     data->state.modified |= SMF_CLIP;
+
+
+     return DFB_OK;
+}
+
 static DFBResult
 IDirectFBSurface_GetGL( IDirectFBSurface   *thiz,
                         IDirectFBGL       **interface )
@@ -2099,7 +2347,7 @@ IDirectFBSurface_Dump( IDirectFBSurface   *thiz,
 
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
 
-     if (!directory || !prefix)
+     if (!directory)
           return DFB_INVARG;
 
      if (!data->area.current.w || !data->area.current.h)
@@ -2114,7 +2362,7 @@ IDirectFBSurface_Dump( IDirectFBSurface   *thiz,
      if (!surface)
           return DFB_DESTROYED;
 
-     return dfb_surface_dump( data->core, surface, directory, prefix );
+     return dfb_surface_dump_buffer( surface, CSBR_FRONT, directory, prefix );
 }
 
 static DFBResult
@@ -2141,6 +2389,68 @@ IDirectFBSurface_ReleaseSource( IDirectFBSurface *thiz )
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
 
      dfb_state_set_source( &data->state, NULL );
+     dfb_state_set_source_mask( &data->state, NULL );
+
+     return DFB_OK;
+}
+
+static DFBResult
+IDirectFBSurface_SetRenderOptions( IDirectFBSurface        *thiz,
+                                   DFBSurfaceRenderOptions  options )
+{
+     DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
+
+     D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
+
+     dfb_state_set_render_options( &data->state, options );
+
+     return DFB_OK;
+}
+
+static DFBResult
+IDirectFBSurface_SetMatrix( IDirectFBSurface *thiz,
+                            const s32        *matrix )
+{
+     DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
+
+     D_DEBUG_AT( Surface, "%s( %p, %p )\n", __FUNCTION__, thiz, matrix );
+
+     if (!matrix)
+          return DFB_INVARG;
+
+     dfb_state_set_matrix( &data->state, matrix );
+
+     return DFB_OK;
+}
+
+static DFBResult
+IDirectFBSurface_SetSourceMask( IDirectFBSurface    *thiz,
+                                IDirectFBSurface    *mask,
+                                int                  x,
+                                int                  y,
+                                DFBSurfaceMaskFlags  flags )
+{
+     DFBResult              ret;
+     DFBPoint               offset = { x, y };
+     IDirectFBSurface_data *mask_data;
+
+     DIRECT_INTERFACE_GET_DATA(IDirectFBSurface)
+
+     D_DEBUG_AT( Surface, "%s( %p, %p, %d,%d, 0x%04x )\n", __FUNCTION__, thiz, mask, x, y, flags );
+
+     if (!mask || flags & ~DSMF_ALL)
+          return DFB_INVARG;
+
+     DIRECT_INTERFACE_GET_DATA_FROM(mask, mask_data, IDirectFBSurface);
+
+     if (!mask_data->surface)
+          return DFB_DESTROYED;
+
+     ret = dfb_state_set_source_mask( &data->state, mask_data->surface );
+     if (ret)
+          return ret;
+
+     dfb_state_set_source_mask_vals( &data->state, &offset, flags );
 
      return DFB_OK;
 }
@@ -2148,6 +2458,7 @@ IDirectFBSurface_ReleaseSource( IDirectFBSurface *thiz )
 /******/
 
 DFBResult IDirectFBSurface_Construct( IDirectFBSurface       *thiz,
+                                      IDirectFBSurface       *parent,
                                       DFBRectangle           *wanted,
                                       DFBRectangle           *granted,
                                       DFBInsets              *insets,
@@ -2155,14 +2466,14 @@ DFBResult IDirectFBSurface_Construct( IDirectFBSurface       *thiz,
                                       DFBSurfaceCapabilities  caps,
                                       CoreDFB                *core )
 {
-     DFBRectangle rect = { 0, 0, surface->width, surface->height };
+     DFBRectangle rect = { 0, 0, surface->config.size.w, surface->config.size.h };
 
      DIRECT_ALLOCATE_INTERFACE_DATA(thiz, IDirectFBSurface)
 
      D_DEBUG_AT( Surface, "%s( %p )\n", __FUNCTION__, thiz );
 
      data->ref = 1;
-     data->caps = caps | surface->caps;
+     data->caps = caps | surface->config.caps;
      data->core = core;
 
      if (dfb_surface_ref( surface )) {
@@ -2170,6 +2481,28 @@ DFBResult IDirectFBSurface_Construct( IDirectFBSurface       *thiz,
           return DFB_FAILURE;
      }
 
+     if (parent && dfb_config->startstop) {
+          IDirectFBSurface_data *parent_data;
+
+          if (parent->AddRef( parent )) {
+               dfb_surface_unref( surface );
+               DIRECT_DEALLOCATE_INTERFACE(thiz);
+               return DFB_FAILURE;
+          }
+
+          DIRECT_INTERFACE_GET_DATA_FROM( parent, parent_data, IDirectFBSurface );
+
+          pthread_mutex_lock( &parent_data->children_lock );
+
+          direct_list_append( &parent_data->children_data, &data->link );
+
+          pthread_mutex_unlock( &parent_data->children_lock );
+
+          data->parent = parent;
+     }
+
+     pthread_mutex_init( &data->children_lock, NULL );
+
      /* The area insets */
      if (insets) {
           data->area.insets = *insets;
@@ -2256,6 +2589,7 @@ DFBResult IDirectFBSurface_Construct( IDirectFBSurface       *thiz,
      thiz->FillTriangle = IDirectFBSurface_FillTriangle;
      thiz->FillRectangles = IDirectFBSurface_FillRectangles;
      thiz->FillSpans = IDirectFBSurface_FillSpans;
+     thiz->FillTriangles = IDirectFBSurface_FillTriangles;
 
      thiz->SetFont = IDirectFBSurface_SetFont;
      thiz->GetFont = IDirectFBSurface_GetFont;
@@ -2271,6 +2605,11 @@ DFBResult IDirectFBSurface_Construct( IDirectFBSurface       *thiz,
      thiz->DisableAcceleration = IDirectFBSurface_DisableAcceleration;
      thiz->ReleaseSource = IDirectFBSurface_ReleaseSource;
 
+     thiz->SetRenderOptions = IDirectFBSurface_SetRenderOptions;
+     thiz->SetMatrix        = IDirectFBSurface_SetMatrix;
+     thiz->SetSourceMask    = IDirectFBSurface_SetSourceMask;
+
+     thiz->MakeSubSurface = IDirectFBSurface_MakeSubSurface;
 
      dfb_surface_attach( surface,
                          IDirectFBSurface_listener, thiz, &data->reaction );
@@ -2300,7 +2639,7 @@ IDirectFBSurface_listener( const void *msg_data, void *ctx )
      }
 
      if (notification->flags & CSNF_SIZEFORMAT) {
-          DFBRectangle rect = { 0, 0, surface->width, surface->height };
+          DFBRectangle rect = { 0, 0, surface->config.size.w, surface->config.size.h };
           
           dfb_rectangle_subtract( &rect, &data->area.insets );
 
@@ -2322,3 +2661,26 @@ IDirectFBSurface_listener( const void *msg_data, void *ctx )
      return RS_OK;
 }
 
+void
+IDirectFBSurface_StopAll( IDirectFBSurface_data *data )
+{
+     if (!dfb_config->startstop)
+          return;
+
+     if (data->children_data) {
+          IDirectFBSurface_data *child;
+
+          pthread_mutex_lock( &data->children_lock );
+
+          direct_list_foreach (child, data->children_data)
+               IDirectFBSurface_StopAll( child );
+
+          pthread_mutex_unlock( &data->children_lock );
+     }
+
+     /* Signal end of sequence of operations. */
+     dfb_state_lock( &data->state );
+     dfb_state_stop_drawing( &data->state );
+     dfb_state_unlock( &data->state );
+}
+
